import logging

from ..constants import OVAL_NAMESPACES, oval_footer, oval_header
from ..xml import ElementTree
from .oval_container import OVALContainer


def _get_xml_element_from_string_shorthand(shorthand):
    valid_oval_xml_string = "{}{}{}".format(oval_header, shorthand, oval_footer).encode(
        "utf-8"
    )
    xml_element = ElementTree.fromstring(valid_oval_xml_string)
    return xml_element.findall("./{%s}def-group/*" % OVAL_NAMESPACES.definition)


def _is_definition_applicable_for_product(definition, product):
    if product is None or definition.is_applicable_for_product(product):
        return True

    logging.info(
        "Definition '{}' is not applicable for product '{}'.".format(
            definition.id_, product
        )
    )
    return False


def _is_definitions_applicable_for_product(definitions_dict, product):
    is_present_applicable_definition = []
    for definition in definitions_dict.values():
        definition.check_affected()

        is_present_applicable_definition.append(
            _is_definition_applicable_for_product(definition, product)
        )
    return any(is_present_applicable_definition)


class OVALShorthand(OVALContainer):
    def _load_element(self, xml_el):
        if xml_el.tag.endswith("definition"):
            self.load_definition(xml_el)
        elif xml_el.tag.endswith("_test"):
            self.load_test(xml_el)
        elif xml_el.tag.endswith("_object"):
            self.load_object(xml_el)
        elif xml_el.tag.endswith("_state"):
            self.load_state(xml_el)
        elif xml_el.tag.endswith("_variable"):
            self.load_variable(xml_el)
        elif xml_el.tag is not ElementTree.Comment:
            logging.warning("Unknown element '{}'\n".format(xml_el.tag))

    @staticmethod
    def _skip_if_is_none(value, component_id):
        return value is None

    def validate(self, product, rule_id):
        if not _is_definitions_applicable_for_product(self.definitions, product):
            return False

        if rule_id is not None and rule_id not in self.definitions:
            raise ValueError(
                (
                    "ERROR: OVAL definition that match the rule ID '{}'"
                    " do not appear in the shorthand."
                ).format(rule_id)
            )
        return True

    def load_shorthand(self, xml_string):
        for xml_el in _get_xml_element_from_string_shorthand(xml_string):
            self._load_element(xml_el)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Script created to help maintainers during the release process by automating Github tasks.
"""

# References:
# - https://developer.github.com/v3/libraries/
# - https://github.com/PyGithub/PyGithub
# - https://pygithub.readthedocs.io/en/latest/github_objects.html
# - https://docs.github.com/en/rest

from datetime import datetime, timedelta, UTC

from github import Github
import argparse
import configparser
import os.path
import re
import subprocess


def get_parameter_from_ini(config_file, section, parameter) -> str:
    config = configparser.ConfigParser()
    try:
        config.read(config_file)
        return config[section][parameter]
    except Exception as e:
        print(f'Error: {e} entry was not found!')


def create_ini_file_template(creds_file):
    try:
        new_creds_file = os.path.expanduser(creds_file)
        with open(new_creds_file, mode='w', encoding='utf-8') as file:
            file.write(
                '[DEFAULT]\n'
                'github_token = <generate your token in https://github.com/settings/tokens>\n'
                '# Include these scopes: "repo:status", "repo_deployment" and "public_repo"')
        print(f'Great! {new_creds_file} was created! Edit it to include your personal token.')
    except Exception as e:
        print(f'Error: {e}')


def get_github_token(creds_file) -> str:
    creds_file = os.path.expanduser(creds_file)
    if os.path.exists(creds_file):
        return get_parameter_from_ini(creds_file, "DEFAULT", "github_token")
    else:
        print(f'{creds_file} file was not found!')
        if get_confirmation(f'Would you like to create the "{creds_file}" file?'):
            create_ini_file_template(creds_file)


def create_github_session(creds_file) -> object:
    token = get_github_token(creds_file)
    if not token:
        print('WARNING: No token found. The API queries are very limited without a token.\n')
    return Github(token)


def get_confirmation(question) -> bool:
    answer = str(input(f'{question} (Y/N): ')).lower().strip()
    return answer[:1] == 'y'


def get_repo_object(session, repo_id) -> object:
    return session.get_repo(repo_id)


def get_repo_root_path() -> str:
    root_path = get_script_path()
    return os.path.dirname(root_path)


def get_script_path() -> str:
    return os.path.dirname(os.path.realpath(__file__))


def show_git_diff():
    subprocess.run(['git', 'diff'], cwd=get_repo_root_path())
    print('\nPlease, review the changes and propose a PR to "master" branch.')


# Repository Branches
def filter_repo_branch_by_name(branches, name) -> object:
    return [branch for branch in branches if branch.name == name]


def get_repo_branches(repo) -> list:
    return repo.get_branches()


def get_repo_stable_branch(repo) -> object:
    branches = get_repo_branches(repo)
    return filter_repo_branch_by_name(branches, "stable")


def get_old_stabilization_branch(repo) -> object:
    branches = get_repo_branches(repo)
    latest_version = get_latest_version(repo)
    branch_name = f'stabilization-v{latest_version}'
    return filter_repo_branch_by_name(branches, branch_name)


def remove_old_stabilization_branch(repo, branch) -> None:
    if get_confirmation(f'Are you sure about removing the "{branch.name}" branch?'):
        # https://github.com/PyGithub/PyGithub/issues/1570
        try:
            branch_ref = repo.get_git_ref(f'heads/{branch.name}')
            branch_ref.delete()
        except Exception as e:
            print(f'Error: {e}')
    else:
        print(f'Aborted! {branch.name} branch not removed.')


# Repository Contributors
def get_contributors_commit_diff(commit) -> str:
    commit_diff = subprocess.run(
        ['git', 'show', '--word-diff', commit, 'Contributors.md'],
        capture_output=True, text=True, cwd=get_repo_root_path())
    return commit_diff.stdout


def get_contributors_last_commit() -> str:
    last_commit = subprocess.run(
        ['git', 'log', '--pretty=format:%h', '-1', '--', 'Contributors.md'],
        capture_output=True, text=True, cwd=get_repo_root_path())
    return last_commit.stdout


def get_contributors_last_update() -> str:
    last_commit = get_contributors_last_commit()
    last_commit_diff = get_contributors_commit_diff(last_commit)

    for line in last_commit_diff.split('\n'):
        # The "generate_contributors.py" creates a commend in the beginning of the file.
        # This comment informs when the file was updated.
        if 'Last Modified:' in line:
            elements = line.split('{+')
            datetime = elements[1].replace('+}', '')
            return datetime


def get_new_contributors(commit) -> str:
    last_commit_diff = get_contributors_commit_diff(commit)
    new_contributors = []

    for line in last_commit_diff.split('\n'):
        if line.startswith('{+'):
            for chars in ['{+', '+}']:
                line = line.replace(chars, '')
            new_contributors.append(line)
    return '\n'.join(new_contributors)


def is_contributors_list_updated(date_string) -> bool:
    date = datetime.strptime(date_string[:16], '%Y-%m-%d %H:%M')
    # As a rule of thumbs, this function consider contributors list
    # updated if not older than 2 weeks.
    two_weeks_back = datetime.now() - timedelta(days=15)
    if date < two_weeks_back:
        return False
    return True


def update_contributors():
    last_update = get_contributors_last_update()
    # Currently, this script considers contributors lists is already updated if not
    # older than 2 weeks.
    if is_contributors_list_updated(last_update):
        print(f'It is all fine, the contributors list was updated in {last_update}')
    else:
        print(f'Contributors list last update was in {last_update}. I can update it for you.')
        utils_scripts_path = get_script_path()
        contributors_script = os.path.join(utils_scripts_path, 'generate_contributors.py')
        subprocess.run(
            [f'PYTHONPATH=. {contributors_script}'], shell=True, cwd=get_repo_root_path())
        show_git_diff()


# Repository Milestones
def has_milestone_version_format(milestone) -> bool:
    regex = re.compile(r'\d\.\d\.\d')
    return regex.match(milestone.title)


def filter_version_milestones(milestones) -> list:
    version_milestones = []
    for milestone in milestones:
        if has_milestone_version_format(milestone):
            version_milestones.append(milestone)
    return version_milestones


def get_repo_milestones(repo) -> list:
    return repo.get_milestones(state="all", direction="desc")


def get_latest_version_milestone(repo) -> object:
    milestones = get_repo_milestones(repo)
    version_milestones = filter_version_milestones(milestones)
    latest_version_milestone = version_milestones[0]
    # It is not ensured the returned milestones list is ordered as expected.
    # This function ensures the last milestone based on titles.
    for milestone in version_milestones:
        if milestone.title > latest_version_milestone.title:
            latest_version_milestone = milestone
    return latest_version_milestone


def get_version_milestone(repo, version) -> object:
    milestones = get_repo_milestones(repo)
    version_milestones = filter_version_milestones(milestones)
    for milestone in version_milestones:
        if milestone.title == version:
            return milestone


def close_milestone(milestone) -> None:
    if milestone.state == "closed":
        print(f'Nice. The "{milestone}" milestone is already closed.')
        return

    if get_confirmation(f'The "{milestone}" milestone should be closed. Ok?'):
        try:
            milestone.edit(milestone.title, state="closed")
        except Exception as e:
            print(f'Error: {e}')
            exit(1)
    else:
        print(f'Aborted! {milestone} milestone not closed.')


def create_repo_milestone(repo, name) -> None:
    if get_version_milestone(repo, name):
        print(f'Great! The "{name}" milestone is already created.')
        return

    latest_release = get_latest_release(repo)
    estimated_release_date = get_next_release_date(latest_release.published_at)
    if get_confirmation(f'Are you sure about creating the "{name}" milestone?'):
        future_release_date = get_next_release_date(estimated_release_date)
        future_stabilization_date = get_next_stabilization_date(future_release_date)
        formatted_date_stabilization = get_date_for_message(future_stabilization_date)
        milestone_description = (
            f'Milestone for the release {name}. '
            f'Stabilization phase starts on {formatted_date_stabilization}')
        try:
            repo.create_milestone(
                title=name, description=milestone_description, due_on=estimated_release_date)
            print(f'Milestone {name} successfully created with the following information:')
            print(f'Description: {milestone_description}')
            print(f'Due on: {estimated_release_date}')
        except Exception as e:
            print(f'Error: {e}')
            exit(1)
    else:
        print(f'Aborted! {name} milestone not created.')


# Repository Releases
def get_repo_releases(repo) -> list:
    return repo.get_releases()


def get_latest_release(repo) -> object:
    releases = get_repo_releases(repo)
    # The API returns the list already ordered by published date.
    return releases[0]


# Repository Versions
def extract_version_from_tag_name(version) -> str:
    return version.split("v")[1]


def get_latest_version(repo) -> str:
    release = get_latest_release(repo)
    return extract_version_from_tag_name(release.tag_name)


# Repository Next Release
def bump_version(current_version, bump_position) -> str:
    # version format example: 0.1.64
    if bump_position == 'minor':
        position = 2
    elif bump_position == 'major':
        position = 1
    else:
        position = 0

    split_version = current_version.split('.')
    bumped = int(split_version[position]) + 1
    split_version[position] = str(bumped)
    return ".".join(split_version)


def bump_master_version(repo):
    content_changed = False
    old_version = get_next_release_version(repo)
    new_version = bump_version(old_version, 'minor')
    old_minor_version = old_version.split('.')[2]
    new_minor_version = new_version.split('.')[2]
    repo_root = get_repo_root_path()
    with open(os.path.join(repo_root, "CMakeLists.txt"), mode='r', encoding='utf-8') as file:
        content = file.readlines()

    for ln, line in enumerate(content):
        if f'set(SSG_PATCH_VERSION {old_minor_version}' in line:
            content[ln] = content[ln].replace(old_minor_version, new_minor_version)
            content_changed = True

    if content_changed:
        with open(os.path.join(repo_root, "CMakeLists.txt"), mode='w', encoding='utf-8') as file:
            file.writelines(content)
        show_git_diff()
    else:
        print('Great! The version in CMakeLists.txt file is already updated.')


def describe_next_release_state(repo) -> str:
    if is_next_release_in_progress(repo):
        return 'In Stabilization Phase'
    else:
        return 'Scheduled'


def get_friday_that_follows(date) -> datetime:
    return date + timedelta((4 - date.weekday()) % 7)


def get_monday_that_follows(date) -> datetime:
    return date + timedelta((0 - date.weekday()) % 7)


def get_next_quarter_first_month(latest_release_date: datetime) -> int:
    last_release_quarter = (latest_release_date.month - 1) // 3 + 1
    if last_release_quarter == 4:
        next_release_quarter = 1
    else:
        next_release_quarter = last_release_quarter + 1
    return (next_release_quarter - 1) * 3 + 1


def get_last_monday_of_month(year: int, month: int) -> datetime:
    if month == 12:
        last_month_day = datetime(year, month, 31)
    else:
        last_month_day = datetime(year, month + 1, 1) - timedelta(days=1)

    days_since_monday = (last_month_day.weekday() - 0) % 7
    return last_month_day - timedelta(days=days_since_monday)


def get_next_stabilization_date(release_date) -> datetime:
    two_weeks_before = release_date - timedelta(weeks=2)
    stabilization_monday = get_monday_that_follows(two_weeks_before)
    return stabilization_monday.date()


def get_next_release_date(latest_release_date: datetime) -> datetime:
    month = get_next_quarter_first_month(latest_release_date)
    now = datetime.now(UTC)

    if month > 9 and latest_release_date <= now:
        year = latest_release_date.year + 1
    else:
        year = latest_release_date.year

    last_monday = get_last_monday_of_month(year, month)
    return get_friday_that_follows(last_monday + timedelta(weeks=1))


def is_next_release_in_progress(repo) -> bool:
    stabilization_branch = get_next_stabilization_branch(repo)
    branches_names = [branch.name for branch in get_repo_branches(repo)]
    if stabilization_branch in branches_names:
        return True
    else:
        return False


def get_next_release_version(repo) -> str:
    current_version = get_latest_version(repo)
    next_version = bump_version(current_version, 'minor')
    return next_version


def get_next_stabilization_branch(repo) -> str:
    next_release = get_next_release_version(repo)
    return f'stabilization-v{next_release}'


# Communication
def get_date_for_message(date) -> datetime:
    return date.strftime("%B %d, %Y")


def get_git_config_username() -> str:
    user_name = subprocess.run(
        ['git', 'config', 'user.name'],
        capture_output=True, text=True, cwd=get_repo_root_path())
    return user_name.stdout


def get_release_highlights(release) -> str:
    highlights = []
    for line in release.body.split('\r\n'):
        if '### Important Highlights' in line:
            continue
        if '###' in line:
            break
        if line:
            highlights.append(line)
    return '\n'.join(highlights)


def get_release_start_message(repo) -> str:
    latest_release = get_latest_release(repo)
    next_release_version = get_next_release_version(repo)
    next_release_date = get_next_release_date(latest_release.published_at)
    date = get_date_for_message(next_release_date)
    branch = get_next_stabilization_branch(repo)

    future_version = bump_version(next_release_version, 'minor')
    future_release_date = get_next_release_date(next_release_date)
    future_date = get_date_for_message(future_release_date)
    future_stabilization_date = get_next_stabilization_date(future_release_date)
    future_date_stabilization = get_date_for_message(future_stabilization_date)

    message_author = get_git_config_username()

    template = f'''
        Subject: stabilization of v{next_release_version}

        Hello all,

        The release of Content version {next_release_version} is scheduled for {date}.
        As part of the release process, a stabilization branch was created.
        Issues and PRs that were not solved were moved to the {future_version} milestone.

        Any bug fixes you would like to include in release {next_release_version} should be
        proposed for the *{branch}* and *master* branches as a measure to avoid
        potential conflicts between these branches.

        The next version, {future_version}, is scheduled to be released on {future_date},
        with the stabilization phase starting on {future_date_stabilization}.

        Regards,

        {message_author}'''
    return template


def get_release_end_message(repo) -> str:
    latest_release = get_latest_release(repo)
    latest_release_url = latest_release.html_url
    highlights = get_release_highlights(latest_release)
    last_commit = get_contributors_last_commit()
    new_contributors = get_new_contributors(last_commit)
    released_version = get_latest_version(repo)
    message_author = get_git_config_username()

    for asset in latest_release.get_assets():
        if asset.content_type == 'application/x-bzip2':
            source_tarball = asset.browser_download_url
        elif asset.content_type == 'application/zip':
            prebuild_zip = asset.browser_download_url
        elif '.tar.bz2.sha512' in asset.name:
            source_tarball_hash = asset.browser_download_url
        else:
            prebuild_zip_hash = asset.browser_download_url

    # It would be more readable if this multiline string would be indented. However, this makes
    # the final text weird, after the "format" substitutions, when highlights and new_contributors
    # have multilines too. So, the readability was compromised in favor of simplicity and user
    # experience. The user only needs to copy and paste, without removing leading spaces.
    template = f'''
Subject: ComplianceAsCode/content v{released_version}

Hello all,

ComplianceAsCode/Content v{released_version} is out.

Some of the highlights of this release are:
{highlights}

Welcome to the new contributors:
{new_contributors}

For full release notes, please have a look at:
{latest_release_url}

Pre-built content: {prebuild_zip}
SHA-512 hash: {prebuild_zip_hash}

Source tarball: {source_tarball}
SHA-512 hash: {source_tarball_hash}

Thank you to everyone who contributed!

Regards,

{message_author}'''
    return template


def print_communication_channels(phase='release') -> None:
    gitter_lnk = 'https://gitter.im/Compliance-As-Code-The/content'
    ghd_lnk = 'https://github.com/ComplianceAsCode/content/discussions/new?category=announcements'
    twitter_lnk = 'https://twitter.com/openscap'
    ssg_mail = 'scap-security-guide@lists.fedorahosted.org'
    openscap_mail = 'open-scap-list@redhat.com'

    print('Please, share the following message in:\n'
          f'* Gitter ({gitter_lnk})\n'
          f'* Github Discussion ({ghd_lnk})\n'
          f'* SCAP Security Guide Mail List ({ssg_mail})')

    if phase == "finish":
        print(f'* OpenSCAP Mail List ({openscap_mail})\n'
              f'* Twitter ({twitter_lnk})')


# Issues and PRs
def filter_items_outdated_milestone(objects_list, milestone) -> list:
    items_old_milestone = []
    for item in objects_list:
        if has_milestone_version_format(item.milestone) and \
                item.milestone.title != milestone.title:
            items_old_milestone.append(item)
    return items_old_milestone


def show_outdated_items(items_to_update):
    count = len(items_to_update)
    print(f'{count} open issues have an outdated milestone. Here are their links:')
    for item in items_to_update:
        print(f'{item.number:5} - {item.html_url:65} - {item.milestone.title} - {item.title}')


def filter_outdated_items(repo, objects_list) -> list:
    latest_milestone = get_latest_version_milestone(repo)
    items_to_update = filter_items_outdated_milestone(objects_list, latest_milestone)
    print('INFO: Please, note that in Github API the Pull Requests are also Issues objects but '
          'with some few differences in properties.')
    if items_to_update:
        show_outdated_items(items_to_update)
        if get_confirmation(
                'Are you sure about updating the milestone in these open issues?'):
            return items_to_update
        else:
            print('Aborted! Milestones not updated in open issues.')
    else:
        print('Great! There is no open issue with an outdated milestone.')
        return []


def get_items_with_milestone(object_list) -> list:
    with_milestone = []
    for item in object_list:
        if item.milestone:
            with_milestone.append(item)
    return with_milestone


def get_open_issues_with_milestone(repo) -> list:
    open_issues = repo.get_issues(state='open', direction='asc')
    return get_items_with_milestone(open_issues)


def get_open_prs_with_milestone(repo) -> list:
    open_prs = repo.get_pulls(state='open', direction='asc')
    return get_items_with_milestone(open_prs)


def update_milestone(repo, object_list) -> None:
    milestone = get_latest_version_milestone(repo)
    for item in object_list:
        try:
            item.edit(milestone=milestone)
        except Exception as e:
            print(f'Error: {e}')


def update_milestone_in_issues(repo, issues_list) -> None:
    outdate_issues = filter_outdated_items(repo, issues_list)
    if outdate_issues:
        update_milestone(repo, outdate_issues)


# Repo Stats
def collect_release_info(repo) -> dict:
    data = dict()
    latest_release = get_latest_release(repo)
    data["latest_release_created"] = latest_release.created_at
    data["latest_release_published"] = latest_release.published_at
    data["latest_release_url"] = latest_release.html_url
    data["latest_version"] = get_latest_version(repo)

    next_release_date = get_next_release_date(latest_release.published_at)
    next_release_remaining_days = next_release_date - datetime.today()
    data["next_release_date"] = next_release_date
    data["next_release_remaining_days"] = next_release_remaining_days.days
    data["next_release_state"] = describe_next_release_state(repo)
    data["next_release_version"] = get_next_release_version(repo)
    next_stabilization_date = get_next_stabilization_date(next_release_date)
    next_stabilization_remaining_days = next_stabilization_date - datetime.today().date()
    data["next_stabilization_date"] = next_stabilization_date
    data["next_stabilization_remaining_days"] = next_stabilization_remaining_days.days

    previous_milestone = get_version_milestone(repo, data["latest_version"])
    data["previous_milestone"] = previous_milestone.title
    data["previous_milestone_closed_issues"] = previous_milestone.closed_issues
    data["previous_milestone_open_issues"] = previous_milestone.open_issues
    data["previous_milestone_total_issues"] = data["previous_milestone_closed_issues"]\
        + data["previous_milestone_open_issues"]

    active_milestone = get_latest_version_milestone(repo)
    data["active_milestone"] = active_milestone.title
    data["active_milestone_closed_issues"] = active_milestone.closed_issues
    data["active_milestone_open_issues"] = active_milestone.open_issues
    data["active_milestone_total_issues"] = data["active_milestone_closed_issues"]\
        + data["active_milestone_open_issues"]
    return data


def print_specific_stat(status, current, total) -> None:
    if current > 0:
        percent = round((current / total) * 100.00, 2)
        print(f'{status:23.23} {current} / {total} = {percent}%')


def print_last_release_stats(data):
    version = data['latest_version']
    created_at = data['latest_release_created']
    published_at = data['latest_release_published']
    url_download = data['latest_release_url']
    closed_issues = data["previous_milestone_closed_issues"]
    total_issues = data["previous_milestone_total_issues"]

    print('Information based on the latest published release:')
    print(f'Current Version:        {version}')
    print(f'Created at:             {created_at}')
    print(f'Published at:           {published_at}')
    print(f'Release Notes:          {url_download}')
    print_specific_stat("Closed Issues/PRs:", closed_issues, total_issues)


def print_next_release_stats(data):
    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    milestone = data['active_milestone']
    next_release_version = data['next_release_version']
    next_release_date = data['next_release_date']
    next_release_days = data['next_release_remaining_days']
    next_release_state = data['next_release_state']
    next_stab_date = data['next_stabilization_date']
    next_stab_days = data['next_stabilization_remaining_days']
    closed_issues = data["active_milestone_closed_issues"]
    total_issues = data["active_milestone_total_issues"]

    print(f'\nNext release information in {now}:')
    print(f'Next Version:           {next_release_version}')
    print(f'Current State:          {next_release_state}')
    print(f'Estimated Release Date: {next_release_date} ({next_release_days} remaining days)')
    print(f'Stabilization Date:     {next_stab_date} ({next_stab_days} remaining days)')
    print(f'Working Milestone:      {milestone}')

    print_specific_stat("Closed Issues/PRs:", closed_issues, total_issues)


# Functions to specific phases of the process
def stats(repo, args):
    try:
        rel_data = collect_release_info(repo)
        print_last_release_stats(rel_data)
        print_next_release_stats(rel_data)
    except Exception as e:
        print(f'Error: {e}')


def cleanup(repo, args) -> None:
    if args.branch:
        branches_to_remove = get_old_stabilization_branch(repo)
        if branches_to_remove:
            for branch in branches_to_remove:
                remove_old_stabilization_branch(repo, branch)
        else:
            print('Great! There is no branch to be removed.')


def release_prep(repo, args) -> None:
    if args.contributors:
        update_contributors()

    stab_branch_name = get_next_stabilization_branch(repo)
    if args.branch:
        print('git checkout master\n'
              'git pull upstream master\n'
              f'git checkout -b {stab_branch_name}\n'
              f'git push -u upstream {stab_branch_name}')

    if args.milestone:
        if is_next_release_in_progress(repo):
            stab_milestone = get_latest_version_milestone(repo)
            next_release_version = get_next_release_version(repo)
            new_milestone = bump_version(next_release_version, 'minor')
            close_milestone(stab_milestone)
            create_repo_milestone(repo, new_milestone)
        else:
            print(f'Milestones can be managed after the "{stab_branch_name}" branch is created.')

    if args.issues:
        issues_list = get_open_issues_with_milestone(repo)
        update_milestone_in_issues(repo, issues_list)

    if args.bump_version:
        bump_master_version(repo)


def release(repo, args) -> None:
    if is_next_release_in_progress(repo):
        if args.tag:
            next_version = get_next_release_version(repo)
            tag_name = f'v{next_version}'
            stab_branch = get_next_stabilization_branch(repo)
            print('Release process starts when a version tag is added to the branch.\n'
                  'It is better to do it manually. But here are the commands you need:\n')
            print(f'git tag {tag_name} {stab_branch}\n'
                  'git push --tags')

        if args.message:
            message = get_release_start_message(repo)
            print_communication_channels('release')
            print(message)
    else:
        print('Please, ensure that all steps in the Release Preparation are concluded.')


def finish(repo, args) -> None:
    if not is_next_release_in_progress(repo):
        if args.message:
            message = get_release_end_message(repo)
            print_communication_channels('finish')
            print(message)
    else:
        print('Please, ensure that all steps in the Release process are concluded.')


def parse_arguments():
    '''Call argparse to process input parameters and return parsed args'''
    parser = argparse.ArgumentParser(
        description="Automate Github tasks included in the Release Process.",
        epilog="Example: release_helper.py -c ~/secrets.ini -r ComplianceAsCode/content stats",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-c', '--creds-file', default='~/secrets.ini',
        help="INI file containing Github token.")
    parser.add_argument(
        '-r', '--repository', action='store', default='ComplianceAsCode/content',
        help="Project repository name.")

    subparsers = parser.add_subparsers(dest='subcmd', required=True)
    subparsers.add_parser(
        'stats', help="Return information about the repository current state.")

    cleanup_parser = subparsers.add_parser(
        'cleanup', help="Cleanup after completing a release.")
    cleanup_parser.add_argument(
        '-b', '--branch', action='store_true',
        help="Remove temporary branch used during stabilization phase.")

    release_prep_parser = subparsers.add_parser(
        'release_prep', help="Prepare the repository for the next release.")
    release_prep_parser.add_argument(
        '-c', '--contributors', action='store_true',
        help="Update the contributors lists. Do it before creating the stabilization branch.")
    release_prep_parser.add_argument(
        '-b', '--branch', action='store_true',
        help="Create the stabilization branch for the next release.")
    release_prep_parser.add_argument(
        '-m', '--milestone', action='store_true',
        help="Create the next milestone and close the current milestone.")
    release_prep_parser.add_argument(
        '-i', '--issues', action='store_true',
        help="Move Open Issues and PRs with a milestone to the next milestone.")
    release_prep_parser.add_argument(
        '-v', '--bump-version', action='store_true',
        help="Bump the project version in *master* branch.")

    release_parser = subparsers.add_parser(
        'release', help="Tasks to be done during the stabilization phase.")
    release_parser.add_argument(
        '-m', '--message', action='store_true',
        help="Prepare a message to communicate the stabilization process has started.")
    release_parser.add_argument(
        '-t', '--tag', action='store_true',
        help="Show commands to properly tag the branch and start the release.")

    finish_parser = subparsers.add_parser(
        'finish', help="Tasks to be done when the release is out.")
    finish_parser.add_argument(
        '-m', '--message', action='store_true',
        help="Prepare a message to communicate the release is out.")
    return parser.parse_args()


def main():
    subcmds = dict(
        stats=stats,
        cleanup=cleanup,
        release_prep=release_prep,
        release=release,
        finish=finish)

    args = parse_arguments()

    try:
        ghs = create_github_session(args.creds_file)
        repo = get_repo_object(ghs, args.repository)
    except Exception as e:
        print(f'Error when getting the repository {args.repository}: {e}')
        exit(1)

    subcmds[args.subcmd](repo, args)


if __name__ == "__main__":
    main()



import ssg.cce as tested_module


def test_is_cce_format_valid():
    icv = tested_module.is_cce_format_valid
    assert icv("CCE-27191-6")
    assert icv("CCE-7223-7")

    assert not icv("not-valid")
    assert not icv("1234-5")
    assert not icv("TBD")
    assert not icv("CCE-TBD")
    assert not icv("CCE-abcde-f")


def test_is_cce_value_valid():
    icv = tested_module.is_cce_value_valid
    assert icv("CCE-27191-6")
    assert icv("CCE-27223-7")

    assert not icv("1234-5")
    assert not icv("12345-6")



#!/usr/bin/python3

import os
import re
import glob

import ssg.build_yaml
import ssg.constants

import tables.table_renderer


class HtmlOutput(tables.table_renderer.TableHtmlOutput):
    TEMPLATE_NAME = "tables/reference_tables_template.html"

    def __init__(self, * args, ** kwargs):
        super(HtmlOutput, self).__init__(* args, ** kwargs)
        self.cached_rules = []

    def _fix_var_sub_in_text(self, text, varname, value):
        return re.sub(
            r'<sub\s+idref="{var}"\s*/>'.format(var=varname),
            r'<abbr title="${var}"><tt>{val}</tt></abbr>'.format(var=varname, val=value), text)

    def _get_eligible_rules(self, refcat):
        filenames = glob.glob(os.path.join(self.rules_root, "*.yml"))
        if self.cached_rules:
            all_rules = self.cached_rules
        else:
            all_rules = [ssg.build_yaml.Rule.from_yaml(f, self.env_yaml) for f in filenames]
            self.cached_rules = all_rules

        rules = []
        for rule in all_rules:
            if refcat in rule.references:
                rules.append(rule)
        return rules

    def process_rules(self, reference):
        super(HtmlOutput, self).process_rules(reference)

        self.template_data["title"] = (
            "{product} rules by {refcat} references"
            .format(product=self.product, refcat=reference.name)
        )


def update_parser(parser):
    pass


def parse_args():
    parser = HtmlOutput.create_parser(
        "Generate HTML table that maps references to rules "
        "using compiled rules as source of data.")
    tables.table_renderer.update_parser(parser)
    update_parser(parser)
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    renderer = HtmlOutput(args.product, args.build_dir, args.verbose)
    reference = ssg.constants.REFERENCES[args.refcategory]
    renderer.process_rules(reference)
    renderer.output_results(args)



#!/usr/bin/python3
import argparse
import csv
import enum
import os.path
import sys

from utils.create_srg_export import DisaStatus

try:
    import ssg.constants
except (ModuleNotFoundError, ImportError):
    sys.stderr.write("Unable to load ssg python modules.\n")
    sys.stderr.write("Hint: run source ./.pyenv.sh\n")
    exit(3)


class Problems(enum.Enum):
    MISSING_REQUIREMENT = "Missing requirement"
    REQUIREMENT_HAS_INVALID_CHARS = "Requirement contains \", &, ', <, or > and is invalid."
    CHECK_BADWORDS = "Check contains should, shall, or please."
    FIX_BADWORDS = "Fix contains should, shall, or please."
    CHECK_BAD_START = "Check starts with ensure or interview"
    FIX_BAD_START = "Fix starts with ensure or interview"
    CHECK_IS_A_FIND = "Check doesn't contain ', this is a finding'"
    INVALID_STATUS = "The status is not valid"
    INVALID_SEVERITY = "Severity is not CAT I, CAT II, or CAT III"


def has_element(search: list, target: str) -> bool:
    return len([elem for elem in search if (elem in target)]) != 0


def has_startswith(search: list, target: str) -> bool:
    return len([elem for elem in search if (target.startswith(elem))]) != 0


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Audits a csv from utils/create_srg_export.py",
                                     epilog="example: ./utils/srg_audit.py build/1642443529_stig_"
                                            "export.csv")
    parser.add_argument('file', type=str, help="A CSV file created by utils/create_srg_export.py")
    return parser.parse_args()


def create_reduction_dicts() -> (dict, dict, dict, dict):
    # The goal these dictionaries is to reduce all values the dictionary to zero,
    # save optional values.
    #
    # Optional values should reduce to 2 if defined or if not defined should remain 3.
    config_dict = {'Requirement': 1,
                   'Vul Discussion': 1,
                   'Status': 1,
                   'Check': 1,
                   'Fix': 1,
                   'Severity': 1,
                   'Mitigation': 0,
                   'Artifact Description': 0,
                   'Status Justification': 3,
                   'CCI': 1,
                   'SRGID': 1}
    meets_dict = {'Requirement': 1,
                  'VulDiscussion': 1,
                  'Status': 1,
                  'Check': 0,
                  'Fix': 0,
                  'Severity': 1,
                  'Mitigation': 0,
                  'Artifact Description': 1,
                  'Status Justification': 1,
                  'CCI': 1,
                  'SRGID': 1}
    does_not_meet = {'Requirement': 1,
                     'Vul Discussion': 1,
                     'Status': 1,
                     'Check': 0,
                     'Fix': 0,
                     'Severity': 1,
                     'Mitigation': 1,
                     'Artifact Description': 0,
                     'Status Justification': 1,
                     'CCI': 1,
                     'SRGID': 1}
    not_applicable = {'Requirement': 1,
                      'Vul Discussion': 0,
                      'Status': 1,
                      'Check': 0,
                      'Fix': 0,
                      'Severity': 1,
                      'Mitigation': 1,
                      'Artifact Description': 0,
                      'Status Justification': 1,
                      'CCI': 1,
                      'SRGID': 1}
    return config_dict, does_not_meet, meets_dict, not_applicable


def check_paths(file: str) -> None:
    if not os.path.exists(file):
        sys.stderr.write("Unable to perform audit.\n")
        sys.stderr.write(f"File not found: {file}\n")
        exit(1)
    if not os.path.isfile(file):
        sys.stderr.write("Unable to perform audit.\n")
        sys.stderr.write(f"Input must be a file: {file}\n")
        exit(2)


def validate_check_fix(results: dict, row: dict, row_id: str) -> None:
    check_invalid = ['shall', 'should', 'please']
    check_invalid_start = ['ensure', 'interview']
    if has_element(check_invalid, row['Fix']):
        results[row_id].append(Problems.FIX_BADWORDS.value)
    if has_element(check_invalid, row['Check']):
        results[row_id].append(Problems.CHECK_BADWORDS.value)
    if has_startswith(check_invalid_start, row['Check']):
        results[row_id].append(Problems.CHECK_BAD_START.value)
    if has_startswith(check_invalid_start, row['Fix']):
        results[row_id].append(Problems.FIX_BAD_START.value)
    if 'this is a finding.' not in row['Check']:
        results[row_id].append(Problems.CHECK_IS_A_FIND.value)


def get_results_dict(row: dict) -> dict:
    config_dict, does_not_meet, meets_dict, not_applicable = create_reduction_dicts()
    if row['Status'] == DisaStatus.AUTOMATED:
        results = config_dict.copy()
    elif row['Status'] == DisaStatus.INHERENTLY_MET:
        results = meets_dict.copy()
    elif row['Status'] == DisaStatus.DOES_NOT_MEET:
        results = does_not_meet.copy()
    elif row['Status'] == DisaStatus.NOT_APPLICABLE:
        results = not_applicable.copy()
    else:
        results = None
    return results


def process_row_results(errors: dict, results: dict, row: dict) -> None:
    for result in results:
        if results[result] == 1:
            errors[get_row_id(row)].append(f'Field {result} is not defined.')
        elif results[result] == -1:
            errors[get_row_id(row)].append(f'Field {result} is defined and should not be.')
        elif results[result] not in [0, 2, 3]:
            raise AttributeError(f"Config data is not defined correctly. "
                                 f"Result = {result}, {results[result]}")


def validate_base_rows(errors: dict, row: dict) -> None:
    if get_row_id(row) not in errors:
        errors[get_row_id(row)] = list()
    if not row['Requirement']:
        errors[get_row_id(row)].append(Problems.MISSING_REQUIREMENT.value)
    req_bad_chars = ['\"', "&", "'", "<", ">"]
    if has_element(req_bad_chars, row['Requirement']):
        errors[get_row_id(row)].append(Problems.REQUIREMENT_HAS_INVALID_CHARS.value)


def set_blank_rows_results(results: dict, row: dict) -> None:
    for col in row:
        col_value = row[col].strip()
        if col_value != '' and col in results:
            results[col] -= 1


def print_results(errors: dict) -> None:
    for result in errors:
        result_errors = errors[str(result)]
        if len(result_errors) > 0:
            print(result)
            for error in result_errors:
                print(f'\t{error}')


def get_row_id(row: dict) -> str:
    if row['STIGID'] is not None and row['STIGID'] != '':
        return row['STIGID']
    else:
        return row['SRGID']


def main():
    args = parse_args()
    check_paths(args.file)

    with open(args.file, 'r') as f:
        reader = csv.DictReader(f)
        errors = dict()
        for row in reader:
            validate_base_rows(errors, row)
            results = get_results_dict(row)
            if row['Status'] == DisaStatus.AUTOMATED:
                validate_check_fix(errors, row, get_row_id(row))
            if row['Status'] not in DisaStatus.STATUSES:
                errors[get_row_id(row)].append(Problems.INVALID_STATUS.value)
                continue
            set_blank_rows_results(results, row)
            process_row_results(errors, results, row)
        print_results(errors)


if __name__ == '__main__':
    main()



"""Abstract and concrete classes for control selection to extract control responses."""

import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List

from ssg.controls import ControlsManager, Control


class ControlSelector(ABC):
    @abstractmethod
    def __init__(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_controls(self) -> List[Control]:
        raise NotImplementedError


class PolicyControlSelector(ControlSelector):
    """Select controls from a policy with optional filtering by level."""

    def __init__(
        self,
        control: str,
        ssg_root: str,
        env_yaml: Dict[str, Any],
        filter_by_level: str = "",
    ) -> None:
        """
        Initialize the PolicyControlSelector.

        Args:
            control: The policy id.
            ssg_root: The path to the root of the ssg directory.
            env_yaml: The environment yaml.
            filter_by_level: Optional level to filter by.
        """
        controls_dir = os.path.join(ssg_root, "controls")
        controls_manager = ControlsManager(controls_dir=controls_dir, env_yaml=env_yaml)
        controls_manager.load()
        if control not in controls_manager.policies:
            raise ValueError(f"Policy {control} not found in controls")

        self.controls: List[Control] = list()
        if filter_by_level:
            policy = controls_manager.policies[control]
            if filter_by_level not in policy.levels_by_id.keys():
                raise ValueError(
                    f"Level {filter_by_level} not found in policy {control}"
                )
            self.controls = controls_manager.get_all_controls_of_level(
                control, filter_by_level
            )
        else:
            self.controls = controls_manager.get_all_controls(control)

    def get_controls(self) -> List[Control]:
        """Get the controls."""
        return self.controls



import json
import os
import pytest
import tempfile

import ssg.build_sce
import ssg.environment
import ssg.products
import ssg.templates


PROJECT_ROOT = os.path.join(os.path.dirname(__file__), "..", "..", "..", )
DATADIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "test_build_sce_data"))
TEST_OUTPUT_DIR = tempfile.mkdtemp()


@pytest.fixture
def scebuilder():
    build_config_yaml_path = os.path.join(
        PROJECT_ROOT, "build", "build_config.yml")
    product_yaml_path = os.path.join(DATADIR, "product.yml")
    env_yaml = ssg.environment.open_environment(
        build_config_yaml_path, product_yaml_path)
    product_yaml = ssg.products.Product(product_yaml_path)
    templates_dir = os.path.join(DATADIR, "templates")
    template_builder = ssg.templates.Builder(
        env_yaml, '', templates_dir,
        '', '', '', None)
    b = ssg.build_sce.SCEBuilder(
        env_yaml, product_yaml, template_builder, TEST_OUTPUT_DIR)
    return b


def test_scebuilder(scebuilder):
    scebuilder.build()

    # Verify that a static SCE check is built
    assert "selinux_state.sh" in os.listdir(TEST_OUTPUT_DIR)
    with open(os.path.join(TEST_OUTPUT_DIR, "selinux_state.sh")) as f:
        contents = f.read()
        assert "$(getenforce) == \"Enforcing\"" in contents

    # Verify that a templated SCE check for a templated rule is built
    assert "package_rsyslog_installed.sh" in os.listdir(TEST_OUTPUT_DIR)
    with open(os.path.join(TEST_OUTPUT_DIR, "package_rsyslog_installed.sh")) as f:
        contents = f.read()
        assert "rpm -q rsyslog" in contents

    # Verify metadata JSON contents
    with open(os.path.join(TEST_OUTPUT_DIR, "metadata.json")) as f:
        metadata = json.load(f)
        assert "selinux_state" in metadata
        assert "package_rsyslog_installed" in metadata



from ssg.utils import parse_template_boolean_value


def preprocess(data, lang):
    data["missing_parameter_pass"] = parse_template_boolean_value(data, parameter="missing_parameter_pass", default_value=False)
    return data



"""
Common functions for processing Jinja2 in SSG
"""

from __future__ import absolute_import
from __future__ import print_function

import os.path
import sys
import jinja2

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

try:
    from shlex import quote as shell_quote
except ImportError:
    from pipes import quote as shell_quote

from .constants import JINJA_MACROS_DIRECTORY
from .utils import (required_key,
                    product_to_name,
                    name_to_platform,
                    product_to_platform,
                    banner_regexify,
                    banner_anchor_wrap,
                    escape_id,
                    escape_regex,
                    escape_yaml_key,
                    sha256
                    )


class MacroError(RuntimeError):
    pass


class AbsolutePathFileSystemLoader(jinja2.BaseLoader):
    """
    AbsolutePathFileSystemLoader is a custom Jinja2 template loader that loads templates from the file system using absolute paths.

    Attributes:
        encoding (str): The encoding used to read the template files. Defaults to 'utf-8'.
    """
    def __init__(self, encoding='utf-8'):
        self.encoding = encoding

    def get_source(self, environment, template):
        """
        Retrieves the source code of a Jinja2 template from the file system.

        Args:
            environment (jinja2.Environment): The Jinja2 environment.
            template (str): The absolute path to the template file.

        Returns:
            tuple: A tuple containing the template contents as a string, the template path, and a
                   function to check if the template file has been updated.

        Raises:
            jinja2.TemplateNotFound: If the template file does not exist or the path is not absolute.
            RuntimeError: If there is an error reading the template file.
        """
        if not os.path.isabs(template):
            raise jinja2.TemplateNotFound(template)

        template_file = jinja2.utils.open_if_exists(template)
        if template_file is None:
            raise jinja2.TemplateNotFound(template)
        try:
            contents = template_file.read().decode(self.encoding)
        except Exception as exc:
            msg = ("Error reading file {template}: {exc}"
                   .format(template=template, exc=str(exc)))
            raise RuntimeError(msg)
        finally:
            template_file.close()

        mtime = os.path.getmtime(template)

        def uptodate():
            try:
                return os.path.getmtime(template) == mtime
            except OSError:
                return False
        return contents, template, uptodate


def _get_jinja_environment(substitutions_dict):
    """
    Initializes and returns a Jinja2 Environment with custom settings and filters.

    This function sets up a Jinja2 Environment with custom block, variable, and comment
    delimiters. It also configures a bytecode cache if specified in the substitutions_dict.
    Additionally, it adds several custom filters to the environment.

    Args:
        substitutions_dict (dict): A dictionary containing configuration options.
            Expected keys include:
            - "jinja2_cache_enabled": A string ("true" or "false") indicating whether bytecode
              caching is enabled.
            - "jinja2_cache_dir": The directory path for storing the bytecode cache
              (required if caching is enabled).

    Returns:
        jinja2.Environment: The configured Jinja2 Environment instance.
    """
    if _get_jinja_environment.env is None:
        bytecode_cache = None
        if substitutions_dict.get("jinja2_cache_enabled") == "true":
            bytecode_cache = jinja2.FileSystemBytecodeCache(
                required_key(substitutions_dict, "jinja2_cache_dir")
            )

        # TODO: Choose better syntax?
        _get_jinja_environment.env = jinja2.Environment(
            block_start_string="{{%",
            block_end_string="%}}",
            variable_start_string="{{{",
            variable_end_string="}}}",
            comment_start_string="{{#",
            comment_end_string="#}}",
            loader=AbsolutePathFileSystemLoader(),
            bytecode_cache=bytecode_cache
        )
        _get_jinja_environment.env.filters['banner_anchor_wrap'] = banner_anchor_wrap
        _get_jinja_environment.env.filters['banner_regexify'] = banner_regexify
        _get_jinja_environment.env.filters['escape_id'] = escape_id
        _get_jinja_environment.env.filters['escape_regex'] = escape_regex
        _get_jinja_environment.env.filters['escape_yaml_key'] = escape_yaml_key
        _get_jinja_environment.env.filters['quote'] = shell_quote
        _get_jinja_environment.env.filters['sha256'] = sha256

    return _get_jinja_environment.env


_get_jinja_environment.env = None


def raise_exception(message):
    raise MacroError(message)


def update_substitutions_dict(filename, substitutions_dict):
    """
    Update the substitutions dictionary with macro definitions from a Jinja2 file.

    This function treats the given filename as a Jinja2 file containing macro definitions.
    It exports definitions that do not start with an underscore (_) into the substitutions_dict,
    which is a dictionary mapping names to macro objects. During the macro compilation process,
    symbols that already exist in substitutions_dict may be used by the new definitions.

    Args:
        filename (str): The path to the Jinja2 file containing macro definitions.
        substitutions_dict (dict): The dictionary to update with new macro definitions.

    Returns:
        None
    """
    template = _get_jinja_environment(substitutions_dict).get_template(filename)
    all_symbols = template.make_module(substitutions_dict).__dict__
    for name, symbol in all_symbols.items():
        if name.startswith("_"):
            continue
        substitutions_dict[name] = symbol


def process_file(filepath, substitutions_dict):
    """
    Process the Jinja file at the given path with the specified substitutions.

    Args:
        filepath (str): The path to the Jinja file to be processed.
        substitutions_dict (dict): A dictionary containing the substitutions to be applied to the template.

    Returns:
        str: The rendered template as a string.

    Note:
        This function does not load the project macros. Use `process_file_with_macros(...)` for that.
    """
    filepath = os.path.abspath(filepath)
    template = _get_jinja_environment(substitutions_dict).get_template(filepath)
    return template.render(substitutions_dict)


def add_python_functions(substitutions_dict):
    """
    Adds predefined Python functions to the provided substitutions dictionary.

    The following functions are added:
    - 'product_to_name': Maps a product identifier to its name.
    - 'name_to_platform': Maps a name to its platform.
    - 'product_to_platform': Maps a product identifier to its platform.
    - 'url_encode': Encodes a URL.
    - 'raise': Raises an exception.
    - 'expand_yaml_path': Expands a YAML path.

    Args:
        substitutions_dict (dict): The dictionary to which the functions will be added.
    """
    substitutions_dict['product_to_name'] = product_to_name
    substitutions_dict['name_to_platform'] = name_to_platform
    substitutions_dict['product_to_platform'] = product_to_platform
    substitutions_dict['url_encode'] = url_encode
    substitutions_dict['raise'] = raise_exception
    substitutions_dict['expand_yaml_path'] = expand_yaml_path


def _load_macros_from_directory(macros_directory, substitutions_dict):
    """
    Helper function to load and update macros from the specified directory.

    Args:
        macros_directory (str): The path to the directory containing macro files.
        substitutions_dict (dict): A dictionary to be augmented with Jinja macros.

    Raises:
        RuntimeError: If there is an error while reading or processing the macro files.
    """
    try:
        for filename in sorted(os.listdir(macros_directory)):
            if filename.endswith(".jinja"):
                macros_file = os.path.join(macros_directory, filename)
                update_substitutions_dict(macros_file, substitutions_dict)
    except Exception as exc:
        msg = ("Error extracting macro definitions from '{1}': {0}"
               .format(str(exc), filename))
        raise RuntimeError(msg)


def _load_macros(macros_directory, substitutions_dict=None):
    """
    Load macros from a specified directory and add them to a substitutions dictionary.

    This function checks if the given macros directory exists, adds Python functions to the
    substitutions dictionary, and then loads macros from the directory into the dictionary.

    Args:
        macros_directory (str): The path to the directory containing macro files.
        substitutions_dict (dict, optional): A dictionary to store the loaded macros.
                                             If None, a new dictionary is created.

    Returns:
        dict: The updated substitutions dictionary containing the loaded macros.

    Raises:
        RuntimeError: If the specified macros directory does not exist.
    """
    if substitutions_dict is None:
        substitutions_dict = dict()

    add_python_functions(substitutions_dict)

    if not os.path.isdir(macros_directory):
        msg = (f"The directory '{macros_directory}' does not exist.")
        raise RuntimeError(msg)

    _load_macros_from_directory(macros_directory, substitutions_dict)

    return substitutions_dict


def load_macros(substitutions_dict=None):
    """
    Augments the provided substitutions_dict with project Jinja macros found in the in
    JINJA_MACROS_DIRECTORY from constants.py.

    Args:
        substitutions_dict (dict, optional): A dictionary to be augmented with Jinja macros.
                                             Defaults to None.

    Returns:
        dict: The updated substitutions_dict containing the Jinja macros.
    """
    return _load_macros(JINJA_MACROS_DIRECTORY, substitutions_dict)


def load_macros_from_content_dir(content_dir, substitutions_dict=None):
    """
    Augments the provided substitutions_dict with project Jinja macros found in a specified
    content directory.

    Args:
        content_dir (str): The base directory containing the 'shared/macros' subdirectory.
        substitutions_dict (dict, optional): A dictionary to be augmented with Jinja macros.
                                             Defaults to None.

    Returns:
        dict: The updated substitutions_dict containing the Jinja macros.
    """
    jinja_macros_directory = os.path.join(content_dir, 'shared', 'macros')
    return _load_macros(jinja_macros_directory, substitutions_dict)


def process_file_with_macros(filepath, substitutions_dict):
    """
    Process a file with Jinja macros.

    This function processes the file located at the given `filepath` using Jinja macros and the
    specified `substitutions_dict`. The `substitutions_dict` is first processed to load any
    macros, and then it is used to substitute values in the file. The function ensures that the
    key 'indent' is not present in the `substitutions_dict`.

    Args:
        filepath (str): The path to the file to be processed.
        substitutions_dict (dict): A dictionary containing the substitutions to be applied to the file.

    Returns:
        str: The processed file content as a string.

    Raises:
        AssertionError: If the key 'indent' is present in `substitutions_dict`.

    See also:
        process_file: A function that processes a file with the given substitutions.
    """
    substitutions_dict = load_macros(substitutions_dict)
    assert 'indent' not in substitutions_dict
    return process_file(filepath, substitutions_dict)


def url_encode(source):
    """
    Encodes a given string into a URL-safe format.

    Args:
        source (str): The string to be URL-encoded.

    Returns:
        str: The URL-encoded string.
    """
    return quote(source)


def expand_yaml_path(path, parameter):
    """
    Expands a dot-separated YAML path into a formatted YAML string.

    Args:
        path (str): The dot-separated path to be expanded.
        parameter (str): An additional parameter to be appended at the end of the path.

    Returns:
        str: A formatted YAML string representing the expanded path.
    """
    out = ""
    i = 0
    for x in path.split("."):
        i += 1
        if i != len(path.split(".")):
            out += i * "  " + x + ":\n"
        elif parameter != "":
            out += i * "  " + x + ":\n"
            i += 1
            out += i * "  " + parameter
        else:
            out += i * "  " + x
    return out


def render_template(data, template_path, output_path, loader):
    """
    Renders a template with the given data and writes the output to a file.

    Args:
        data (dict): The data to be used in the template rendering.
        template_path (str): The path to the template file.
        output_path (str): The path where the rendered output will be written.
        loader (jinja2.BaseLoader): The Jinja2 loader to use for loading templates.

    Returns:
        None
    """
    env = _get_jinja_environment(dict())
    env.loader = loader
    result = process_file(template_path, data)
    with open(output_path, "wb") as f:
        f.write(result.encode('utf8', 'replace'))



#!/usr/bin/python3

import argparse
import os
import re
import sys
from collections import defaultdict
from typing import Dict, Iterator, List, Pattern

RefsFoundType = "Dict[str, List[str]]"

rx_quoted = re.compile(r'^([\'"])(.*?)\1$')


def recursive_get_rule_files() -> Iterator[str]:
    """Recursive walk whole tree and get rule.yml paths."""
    for root, _, filenames in os.walk("."):
        for filename in filenames:
            full_path = os.path.join(root, filename)
            if os.path.basename(full_path) == "rule.yml":
                yield os.path.normpath(full_path)


def find_rule_file_refs(rule_file: str, rx_ref_line: Pattern[str]) -> Iterator[str]:
    """Find refs per file, either as is or quoted."""
    with open(rule_file, "r") as fd:
        for line in fd:
            match = rx_ref_line.match(line)
            if match:
                found = match.group(1)
                match_quoted = rx_quoted.match(found)
                if match_quoted:
                    found = match_quoted.group(2)
                yield found


def find_refs(ref: str) -> RefsFoundType:
    """Find refs from all rule files."""
    rx_ref_line = re.compile(r"^\s*%s(?:@\w*)?:\s*(.*?)\s*$" % (re.escape(ref)))

    refs_found = defaultdict(list)
    for rule_file in recursive_get_rule_files():
        for found in find_rule_file_refs(rule_file, rx_ref_line):
            refs_found[found].append(rule_file)

    return refs_found


def print_refs(ref: str, refs_found: RefsFoundType) -> bool:
    """
    Print refs and files if ref is found in two or more files.

    Print info into stderr about refs and files if so.
    Returns: True if none found, False if found.
    """
    okay = True

    for dup in sorted(refs_found.keys()):
        filenames = refs_found[dup]
        if len(filenames) > 1:
            okay = False
            print("", file=sys.stderr)
            print(f"{ref} {dup} is included in files: ", file=sys.stderr)
            for filename in sorted(filenames):
                print(f" - {filename}", file=sys.stderr)

    return okay


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("root_ssg_directory", help="Path to root of ssg git repository")
    parser.add_argument("reference", help="Reference to search in rule.yml files")
    return parser.parse_args()


def main() -> None:
    args = parse_args()
    root_dir = args.root_ssg_directory
    ref = args.reference

    if not os.path.isdir(root_dir):
        print(
            f"ERROR: root_ssg_directory '{root_dir}' is not a directory",
            file=sys.stderr,
        )
        sys.exit(7)

    os.chdir(root_dir)

    refs_found = find_refs(ref)
    okay = print_refs(ref, refs_found)
    sys.exit(0 if okay else 1)


if __name__ == "__main__":
    main()



from ssg.utils import parse_template_boolean_value


def preprocess(data, lang):
    data["arg_negate"] = parse_template_boolean_value(data, parameter="arg_negate", default_value=False)
    data["arg_is_regex"] = parse_template_boolean_value(data, parameter="arg_is_regex", default_value=False)
    return data



"""
Common functions for processing CCE (Common Configuration Enumeration) in SSG
"""

import re
import random
import os


CCE_POOLS = dict()


class CCEFile:
    """
    A class to represent a CCE (Common Configuration Enumeration) file.

    Attributes:
        project_root (str): The root directory of the project.
    """
    def __init__(self, project_root=None):
        if not project_root:
            project_root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "..")
        self.project_root = project_root

    @property
    def absolute_path(self):
        raise NotImplementedError()

    def line_to_cce(self, line):
        return line

    def line_isnt_cce(self, cce, line):
        return line != cce

    def read_cces(self):
        """
        Reads the CCEs (Common Configuration Enumeration) from a file and validates them.

        This method opens the file specified by `self.absolute_path`, reads its contents
        line by line, and checks each line to ensure it is a valid CCE value. If an invalid CCE
        is detected, a RuntimeError is raised with an appropriate error message.

        Returns:
            list: A list of valid CCE values read from the file.

        Raises:
            RuntimeError: If an invalid CCE value is detected in the file.
        """
        with open(self.absolute_path, "r") as f:
            cces = f.read().splitlines()
        for cce in cces:
            if not is_cce_value_valid(cce):
                msg = (
                    "Invalid CCE detected in {cce_path}: {cce}"
                    .format(cce=cce, cce_path=self.absolute_path))
                raise RuntimeError(msg)
        return cces

    def remove_cce_from_file(self, cce):
        """
        Removes lines containing the specified CCE (Common Configuration Enumeration) from a file.

        Args:
            cce (str): The CCE identifier to be removed from the file.

        The method reads the file, filters out lines containing the specified CCE,
        and writes the remaining lines back to the file.
        """
        file_lines = self.read_cces()
        lines_except_cce = [
            line for line in file_lines
            if self.line_isnt_cce(cce, line)
        ]
        with open(self.absolute_path, "w") as f:
            f.write("\n".join(lines_except_cce) + "\n")

    def random_cce(self):
        """
        Selects a random CCE (Common Configuration Enumeration) from the list of CCEs.

        Reads the list of CCEs, shuffles them randomly, and returns the first CCE
        from the shuffled list after stripping any leading or trailing whitespace.

        Returns:
            str: A randomly selected and stripped CCE.
        """
        cces = self.read_cces()
        random.shuffle(cces)
        return cces[0].strip()


class RedhatCCEFile(CCEFile):
    """
    RedhatCCEFile is a subclass of CCEFile that represents a file containing
    Red Hat Common Configuration Enumeration (CCE) data.

    Properties:
        absolute_path (str): The absolute path to the Red Hat CCE file, which is located in the
                             "shared/references" directory within the project root and named "cce-redhat-avail.txt".
    """
    @property
    def absolute_path(self):
        return os.path.join(self.project_root, "shared", "references", "cce-redhat-avail.txt")


class SLE12CCEFile(CCEFile):
    """
    SLE12CCEFile is a subclass of CCEFile that represents a file containing
    SLE12 Common Configuration Enumeration (CCE) data.

    This class provides a property to get the absolute path of the SLE12 CCE file.

    Properties:
        absolute_path (str): The absolute path to the SLE12 CCE file, which is located in the
                             "shared/references" directory.
    """
    @property
    def absolute_path(self):
        return os.path.join(self.project_root, "shared", "references", "cce-sle12-avail.txt")


class SLE15CCEFile(CCEFile):
    """
    SLE15CCEFile is a subclass of CCEFile that represents a file containing
    SLE15 Common Configuration Enumeration (CCE) data.

    This class provides a property to get the absolute path of the SLE12 CCE file.

    Properties:
        absolute_path (str): The absolute path to the SLE15 CCE file, which is located in the
                             "shared/references" directory.
    """
    @property
    def absolute_path(self):
        return os.path.join(self.project_root, "shared", "references", "cce-sle15-avail.txt")


CCE_POOLS["redhat"] = RedhatCCEFile
CCE_POOLS["sle12"] = SLE12CCEFile
CCE_POOLS["sle15"] = SLE15CCEFile


def is_cce_format_valid(cceid):
    """
    Check if the given CCE ID is in a valid format.

    A valid CCE ID must be in one of the following formats:
    - 'CCE-XXXX-X'
    - 'CCE-XXXXX-X'

    where each 'X' is a digit, and the final 'X' is a check-digit.

    Args:
        cceid (str): The CCE ID to validate.

    Returns:
        bool: True if the CCE ID is in a valid format, False otherwise.
    """
    match = re.match(r'^CCE-\d{4,5}-\d$', cceid)
    return match is not None


def is_cce_value_valid(cceid):
    """
    Validates a CCE (Common Configuration Enumeration) identifier using Luhn's algorithm.

    This function removes non-digit characters from the CCE identifier and then applies
    Luhn's algorithm to determine if the identifier is valid.

    Args:
        cceid (str): The CCE identifier to validate.

    Returns:
        bool: True if the CCE identifier is valid, False otherwise.

    References:
        For context, see:
        https://github.com/ComplianceAsCode/content/issues/3044#issuecomment-420844095
    """
    # concat(substr ... , substr ...) -- just remove non-digit characters.
    # Since we've already validated format, this hack suffices:
    cce = re.sub(r'(CCE|-)', '', cceid)

    # The below is an implementation of Luhn's algorithm as this is what the
    # XPath code does.

    # First, map string numbers to integers. List cast is necessary to be able
    # to index it.
    digits = list(map(int, cce))

    # Even indices are doubled. Coerce to list for list addition. However,
    # XPath uses 1-indexing so "evens" and "odds" are swapped from Python.
    # We handle both the idiv and the mod here as well; note that we only
    # hvae to do this for evens: no single digit is above 10, so the idiv
    # always returns 0 and the mod always returns the original number.
    evens = list(map(lambda i: (i*2)//10 + (i*2) % 10, digits[-2::-2]))
    odds = digits[-1::-2]

    # The checksum value is now the sum of the evens and the odds.
    value = sum(evens + odds) % 10

    # Valid CCE <=> value == 0
    return value == 0



#!/usr/bin/python3

from __future__ import print_function

from glob import glob
import os

import ssg.build_yaml
import ssg.controls
import ssg.yaml
import ssg.jinja
import template_renderer
from rendering import common


class HtmlOutput(template_renderer.Renderer):
    TEMPLATE_NAME = "rendering/rules-template.html"

    def _get_all_compiled_profiles(self):
        compiled_profiles = glob(os.path.join(self.built_content_path, "profiles", "*.profile"))
        profiles = []
        for p in compiled_profiles:
            profiles.append(ssg.build_yaml.Profile.from_yaml(p, self.env_yaml))
        return profiles

    def _set_rule_profiles_membership(self, rule, profiles):
        rule.in_profiles = [p for p in profiles if rule.id_ in p.selected]

    def process_rules(self, rule_files):
        rules = []
        profiles = self._get_all_compiled_profiles()

        for r_file in rule_files:
            rule = ssg.build_yaml.Rule.from_yaml(r_file, self.env_yaml)
            self._set_rule_relative_definition_location(rule)
            self._set_rule_profiles_membership(rule, profiles)
            common.resolve_var_substitutions_in_rule(rule)
            rules.append(rule)

        self.template_data["rules"] = rules


def parse_args():
    parser = HtmlOutput.create_parser(
        "Pass a list of rule YAMLs, and the script will "
        "render their summary as an HTML along with links to the usage of these rules in profiles "
        "and with links to the upstream rule source.")
    parser.add_argument(
        "rule", metavar="FILENAME", nargs="+", help="The rule YAML files")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    renderer = HtmlOutput(args.product, args.build_dir)
    renderer.process_rules(args.rule)
    renderer.output_results(args)



import collections
import copy
import os
import pytest
import xml.etree.ElementTree as ET

import ssg.id_translate
from ssg.oval_object_model import load_oval_document
from ssg.constants import XCCDF12_NS, oval_namespace, ocil_namespace

DATADIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
oval_unix_namespace = oval_namespace + "#unix"


def test__split_namespace():
    sn = ssg.id_translate._split_namespace

    ns, n = sn("{oval}board")
    assert ns == "oval"
    assert n == "board"

    ns, n = sn("{oval#magic}board")
    assert ns == "oval"
    assert n == "board"

    ns, n = sn("nonamespace")
    assert not ns
    assert n == "nonamespace"

    ns, n = sn("{}emptynamespace")
    assert not ns
    assert n == "emptynamespace"


def test_tagname_to_abbrev():
    tag_expectations = {
        "definition": "def",
        "criteria": "crit",
        "test": "tst",
        "object": "obj",
        "state": "ste",
        "variable": "var",
    }
    for tag, expectation in tag_expectations.items():
        real = ssg.id_translate._tagname_to_abbrev(
            "{%s}%s" % (oval_namespace, tag))
        assert real == expectation


def test_tagname_to_abbrev_unknown_ns():
    with pytest.raises(RuntimeError):
        unknown_name_space = "http://world.wide.web/some/unknown/foo.php"
        ssg.id_translate._tagname_to_abbrev(
            "{%s}rationale" % unknown_name_space)


def test_tagname_to_abbrev_unknown_tag():
    with pytest.raises(KeyError):
        ssg.id_translate._tagname_to_abbrev("{%s}candy" % oval_namespace)
    with pytest.raises(KeyError):
        ssg.id_translate._tagname_to_abbrev("{%s}pizza" % ocil_namespace)


@pytest.fixture
def idtranslator():
    return ssg.id_translate.IDTranslator("ssg")


def test_idtranslator_generate_id(idtranslator):
    genid = idtranslator.generate_id(
        "{%s}object" % oval_namespace, "my_terrible_object")
    assert genid == "oval:ssg-my_terrible_object:obj:1"
    genid = idtranslator.generate_id(
        "{%s}question" % ocil_namespace, "my_terrible_face")
    assert genid == "ocil:ssg-my_terrible_face:question:1"


@pytest.fixture
def oval_tree():
    draft_oval_path = os.path.join(DATADIR, "draft_oval.xml")
    return ET.parse(draft_oval_path).getroot()


def _parse_interesting_ids(oval_tree):
    InterestingIDs = collections.namedtuple(
        "InterestingIDs", [
            "definition_id", "criterion_test_ref", "test_id",
            "object_ref", "object_id", "filter_ref", "state_id"])
    ns = {"oval": oval_namespace, "oval-unix": oval_namespace + "#unix"}
    definition_id = oval_tree.find(
        "oval:definitions/oval:definition", ns).get("id")
    criterion_test_ref = oval_tree.find(
        "oval:definitions/oval:definition/oval:criteria/oval:criterion",
        ns).get("test_ref")
    test_id = oval_tree.find(
        "oval:tests/oval-unix:file_test", ns).get("id")
    object_ref = oval_tree.find(
        "oval:tests/oval-unix:file_test/oval-unix:object",
        ns).get("object_ref")
    object_id = oval_tree.find(
        "oval:objects/oval-unix:file_object",
        ns).get("id")
    filter_ref = oval_tree.find(
        "oval:objects/oval-unix:file_object/oval:filter", ns).text
    state_id = oval_tree.find("oval:states/oval-unix:file_state", ns).get("id")
    iids = InterestingIDs(
        definition_id=definition_id, criterion_test_ref=criterion_test_ref,
        test_id=test_id, object_ref=object_ref, object_id=object_id,
        filter_ref=filter_ref, state_id=state_id)
    return iids


def test_idtranslator_translate_oval_ids(idtranslator, oval_tree):
    expected_definition_id = "oval:ssg-kerberos_disable_no_keytab:def:1"
    expected_test_id = "oval:ssg-test_kerberos_disable_no_keytab:tst:1"
    expected_object_id = "oval:ssg-obj_kerberos_disable_no_keytab:obj:1"
    expected_state_id = "oval:ssg-filter_ssh_key_owner_root:ste:1"
    oval_document = load_oval_document(oval_tree)
    new_oval_document = idtranslator.translate_oval_document(oval_document)
    new_tree = new_oval_document.get_xml_element()
    real = _parse_interesting_ids(new_tree)
    assert real.definition_id == expected_definition_id
    assert real.criterion_test_ref == expected_test_id
    assert real.criterion_test_ref == real.test_id
    assert real.test_id == expected_test_id
    assert real.object_ref == expected_object_id
    assert real.object_ref == real.object_id
    assert real.object_id == expected_object_id
    assert real.filter_ref == expected_state_id
    assert real.filter_ref == real.state_id
    assert real.state_id == expected_state_id


def test_idtranslator_translate_oval_ids_using_oval_object_model(idtranslator, oval_tree):
    expected_definition_id = "oval:ssg-kerberos_disable_no_keytab:def:1"
    expected_test_id = "oval:ssg-test_kerberos_disable_no_keytab:tst:1"
    expected_object_id = "oval:ssg-obj_kerberos_disable_no_keytab:obj:1"
    expected_state_id = "oval:ssg-filter_ssh_key_owner_root:ste:1"
    new_tree = idtranslator.translate(oval_tree)
    real = _parse_interesting_ids(new_tree)
    assert real.definition_id == expected_definition_id
    assert real.criterion_test_ref == expected_test_id
    assert real.criterion_test_ref == real.test_id
    assert real.test_id == expected_test_id
    assert real.object_ref == expected_object_id
    assert real.object_ref == real.object_id
    assert real.object_id == expected_object_id
    assert real.filter_ref == expected_state_id
    assert real.filter_ref == real.state_id
    assert real.state_id == expected_state_id


def test_idtranslator_translate_oval_xmldiff(idtranslator, oval_tree):
    xmldiff_main = pytest.importorskip("xmldiff.main")
    xmldiff_actions = pytest.importorskip("xmldiff.actions")
    LET = pytest.importorskip("lxml.etree")
    old = LET.fromstring(ET.tostring(oval_tree))
    new = LET.fromstring(ET.tostring(idtranslator.translate(oval_tree)))
    inverted_new_nsmap = {v: k for k, v in old.nsmap.items()}
    o = inverted_new_nsmap[oval_namespace]
    ou = inverted_new_nsmap[oval_unix_namespace]
    diff = set(xmldiff_main.diff_trees(old, new))
    uadefid = xmldiff_actions.UpdateAttrib(
        node='/{o}:oval_definitions/{o}:definitions/{o}:definition[1]'.format(o=o),
        name='id',
        value='oval:ssg-kerberos_disable_no_keytab:def:1')
    assert uadefid in diff
    diff.remove(uadefid)
    uaftestid = xmldiff_actions.UpdateAttrib(
        node='/{o}:oval_definitions/{o}:tests/{ou}:file_test[1]'.format(o=o, ou=ou),
        name='id',
        value='oval:ssg-test_kerberos_disable_no_keytab:tst:1')
    assert uaftestid in diff
    diff.remove(uaftestid)
    uafoid = xmldiff_actions.UpdateAttrib(
        node='/{o}:oval_definitions/{o}:objects/{ou}:file_object[1]'.format(o=o, ou=ou),
        name='id',
        value='oval:ssg-obj_kerberos_disable_no_keytab:obj:1')
    assert uafoid in diff
    diff.remove(uafoid)
    uafsid = xmldiff_actions.UpdateAttrib(
        node='/{o}:oval_definitions/{o}:states/{ou}:file_state[1]'.format(o=o, ou=ou),
        name='id',
        value='oval:ssg-filter_ssh_key_owner_root:ste:1')
    assert uafsid in diff
    diff.remove(uafsid)
    uaftoref = xmldiff_actions.UpdateAttrib(
        node=(
            '/{o}:oval_definitions/{o}:tests/{ou}:file_test/'
            '{ou}:object[1]'.format(ou=ou, o=o)),
        name='object_ref',
        value='oval:ssg-obj_kerberos_disable_no_keytab:obj:1')
    assert uaftoref in diff
    diff.remove(uaftoref)
    utfofilter = xmldiff_actions.UpdateTextIn(
        node=(
            '/{o}:oval_definitions/{o}:objects/{ou}:file_object/'
            '{o}:filter[1]'.format(o=o, ou=ou)),
        text='oval:ssg-filter_ssh_key_owner_root:ste:1')
    assert utfofilter in diff
    diff.remove(utfofilter)
    uacrittref = xmldiff_actions.UpdateAttrib(
        node=(
            '/{o}:oval_definitions/{o}:definitions/{o}:definition/'
            '{o}:criteria/{o}:criterion[1]'.format(o=o)),
        name='test_ref',
        value='oval:ssg-test_kerberos_disable_no_keytab:tst:1')
    assert uacrittref in diff
    diff.remove(uacrittref)
    assert diff == set()


def test_idtranslator_translate_oval_store_defname(idtranslator, oval_tree):
    new_tree = idtranslator.translate(oval_tree, True)
    definition_el = new_tree.find(
        "{%s}definitions/{%s}definition" % (oval_namespace, oval_namespace))
    ref_els = definition_el.findall(
        "{%s}metadata/{%s}reference" % (oval_namespace, oval_namespace))
    assert len(ref_els) == 1
    ref_el = ref_els[0]
    assert ref_el is not None
    assert len(ref_el.attrib) == 2
    assert ref_el.get("ref_id") == "kerberos_disable_no_keytab"
    assert ref_el.get("source") == "ssg"


@pytest.fixture
def ocil_tree():
    draft_ocil_path = os.path.join(DATADIR, "draft_ocil.xml")
    return ET.parse(draft_ocil_path).getroot()


def test_idtranslator_translate_ocil(idtranslator, ocil_tree):
    questionnaire_id = "ocil:ssg-service_cockpit_disabled_ocil:questionnaire:1"
    test_action_id = "ocil:ssg-service_cockpit_disabled_action:testaction:1"
    question_id = "ocil:ssg-service_cockpit_disabled_question:question:1"
    new_tree = idtranslator.translate(ocil_tree)
    assert new_tree is not None
    assert new_tree.tag == "{%s}ocil" % ocil_namespace
    questionnaires_els = new_tree.findall(
        "{%s}questionnaires" % ocil_namespace)
    assert len(questionnaires_els) == 1
    questionnaires_el = questionnaires_els[0]
    assert len(questionnaires_el.attrib) == 0
    questionnaire_els = questionnaires_el.findall(
        "{%s}questionnaire" % ocil_namespace)
    assert len(questionnaire_els) == 1
    questionnaire_el = questionnaires_el[0]
    assert len(questionnaire_el.attrib) == 1
    assert questionnaire_el.get("id") == questionnaire_id
    test_action_ref_els = questionnaire_el.findall(
        "{%s}actions/{%s}test_action_ref" % (ocil_namespace, ocil_namespace))
    assert len(test_action_ref_els) == 1
    test_action_ref_el = test_action_ref_els[0]
    assert len(test_action_ref_el.attrib) == 0
    assert test_action_ref_el.text == test_action_id
    test_actions_els = new_tree.findall("{%s}test_actions" % ocil_namespace)
    assert len(test_actions_els) == 1
    test_actions_el = test_actions_els[0]
    assert len(test_actions_el.attrib) == 0
    boolean_question_test_action_els = test_actions_el.findall(
        "{%s}boolean_question_test_action" % ocil_namespace)
    assert len(boolean_question_test_action_els) == 1
    boolean_question_test_action_el = boolean_question_test_action_els[0]
    assert len(boolean_question_test_action_el.attrib) == 2
    assert boolean_question_test_action_el.get("id") == test_action_id
    assert boolean_question_test_action_el.get("question_ref") == question_id
    questions_els = new_tree.findall("{%s}questions" % ocil_namespace)
    assert len(questions_els) == 1
    questions_el = questions_els[0]
    boolean_question_els = questions_el.findall(
        "{%s}boolean_question" % ocil_namespace)
    assert len(boolean_question_els) == 1
    boolean_question_el = boolean_question_els[0]
    assert len(boolean_question_el.attrib) == 1
    assert boolean_question_el.get("id") == question_id



"""
Common functions for processing OVAL in SSG
"""

from __future__ import absolute_import
from __future__ import print_function

import sys
import os

from .constants import oval_footer as footer
from .constants import oval_namespace as ovalns
from .xml import ElementTree as ET
from .xml import oval_generated_header

from .jinja import process_file_with_macros
from .products import _get_implied_properties


ASSUMED_OVAL_VERSION_STRING = "5.11"
# globals, to make recursion easier in case we encounter extend_definition
try:
    ET.register_namespace("oval", ovalns)
except AttributeError:
    # Legacy Python 2.6 fix, see e.g.
    # https://www.programcreek.com/python/example/57552/xml.etree.ElementTree._namespace_map
    from xml.etree import ElementTree as ET
    ET._namespace_map[ovalns] = "oval"


def applicable_platforms(oval_file, oval_version_string=None):
    """
    Returns the applicable platforms for a given OVAL file.

    This function processes an OVAL file to extract the platforms it applies to. It uses a
    specified OVAL version string or a default version if none is provided. The function
    constructs an XML tree from the OVAL file and extracts platform information from it.

    Args:
        oval_file (str): The path to the OVAL file to be processed.
        oval_version_string (str, optional): The OVAL version string to be used.
                                             If not provided, a default version is used.

    Returns:
        list: A list of platforms that the OVAL file applies to.

    Raises:
        Exception: If there is an error while parsing the OVAL file.
    """
    platforms = []

    if not oval_version_string:
        oval_version_string = ASSUMED_OVAL_VERSION_STRING
    header = oval_generated_header("applicable_platforms", oval_version_string, "0.0.1")

    oval_version_list = [int(num) for num in oval_version_string.split(".")]
    subst_dict = dict(target_oval_version=oval_version_list)

    oval_filename_components = oval_file.split(os.path.sep)
    if len(oval_filename_components) > 3:
        subst_dict["rule_id"] = oval_filename_components[-3]
    else:
        msg = "Unable to get rule ID from OVAL path '{path}'".format(path=oval_file)
        print(msg, file=sys.stderr)

    subst_dict = _get_implied_properties(subst_dict)
    subst_dict['target_oval_version'] = [999, 999.999]

    body = process_file_with_macros(oval_file, subst_dict)

    try:
        oval_tree = ET.fromstring(header + body + footer)
    except Exception as e:
        msg = "Error while loading " + oval_file
        print(msg, file=sys.stderr)
        raise e

    element_path = "./{%s}def-group/{%s}definition/{%s}metadata/{%s}affected/{%s}platform"
    element_ns_path = element_path % (ovalns, ovalns, ovalns, ovalns, ovalns)
    for node in oval_tree.findall(element_ns_path):
        platforms.append(node.text)

    return platforms


def parse_affected(oval_contents):
    """
    Returns the tuple (start_affected, end_affected, platform_indents) for the passed OVAL file contents.

    Args:
    oval_contents (list of str): The contents of the OVAL file, where each element is a line from
                                 the file.

    Returns:
        tuple: A tuple containing:
            - start_affected (int): The line number of the starting <affected> tag.
            - end_affected (int): The line number of the closing </affected> tag.
            - platform_indents (str): The indenting characters before the contents of the
              <affected> element.

    Raises:
        ValueError: If the OVAL file does not contain a single <affected> element, if the start
                    tag is after the end tag, or if the tags contain other elements.
    """
    start_affected = list(filter(lambda x: "<affected" in oval_contents[x],
                                 range(0, len(oval_contents))))
    if len(start_affected) != 1:
        raise ValueError("OVAL file does not contain a single <affected> "
                         "element; counted %d in:\n%s\n\n" %
                         (len(start_affected), "\n".join(oval_contents)))

    start_affected = start_affected[0]

    end_affected = list(filter(lambda x: "</affected" in oval_contents[x],
                               range(0, len(oval_contents))))
    if len(end_affected) != 1:
        raise ValueError("Malformed OVAL file does not contain a single "
                         "closing </affected>; counted %d in:\n%s\n\n" %
                         (len(start_affected), "\n".join(oval_contents)))
    end_affected = end_affected[0]

    if start_affected >= end_affected:
        raise ValueError("Malformed OVAL file: start affected tag begins "
                         "on the same line or after ending affected tag: "
                         "start:%d vs end:%d:\n%s\n\n" %
                         (start_affected, end_affected, oval_contents))

    # Validate that start_affected contains only a starting <affected> tag;
    # otherwise, using this information to update the <platform> subelements
    # would fail.
    start_line = oval_contents[start_affected]
    start_line = start_line.strip()

    if not start_line.startswith('<affected'):
        raise ValueError("Malformed OVAL file: line with starting affected "
                         "tag contains other elements: line:%s\n%s\n\n" %
                         (start_line, oval_contents))
    if '<' in start_line[1:]:
        raise ValueError("Malformed OVAL file: line with starting affected "
                         "tag contains other elements: line:%s\n%s\n\n" %
                         (start_line, oval_contents))

    # Validate that end_affected contains only an ending </affected> tag;
    # otherwise, using this information to update the <platform> subelements
    # would fail.
    end_line = oval_contents[end_affected]
    end_line = end_line.strip()

    if not end_line.startswith('</affected>'):
        raise ValueError("Malformed OVAL file: line with ending affected "
                         "tag contains other elements: line:%s\n%s\n\n" %
                         (end_line, oval_contents))
    if '<' in end_line[1:]:
        raise ValueError("Malformed OVAL file: line with ending affected "
                         "tag contains other elements: line:%s\n%s\n\n" %
                         (end_line, oval_contents))

    indents = ""
    if start_affected+1 == end_affected:
        # Since the affected element is present but empty, the indents should
        # be two more spaces than that of the starting <affected> element.
        start_index = oval_contents[start_affected].index('<')
        indents = oval_contents[start_affected][0:start_index]
        indents += "  "
    else:
        # Otherwise, grab the indents off the next line unmodified, as this is
        # likely a platform element tag. We don't validate here that this is
        # indeed the case, as other parts of the build infrastructure will
        # validate this for us.
        start_index = oval_contents[start_affected+1].index('<')
        indents = oval_contents[start_affected+1][0:start_index]

    return start_affected, end_affected, indents



#!/usr/bin/python3
import argparse
import os
from utils.template_renderer import render_template
from ssg.xml import ElementTree as ET
from ssg.xml import determine_xccdf_tree_namespace


TABLE_DIR = os.path.join(os.path.dirname(__file__), "tables")
STIG_TEMPLATE = os.path.join(TABLE_DIR, "stig_template.html")


def parse_args():
    parser = argparse.ArgumentParser(
        description="Create a HTML STIG table")
    parser.add_argument("input", help="Input XCCDF file")
    parser.add_argument("output", help="Output HTML file")
    return parser.parse_args()


def get_stats(root, ns):
    total = 0
    missing = 0
    missing_rules = []
    for rule in root.findall("xccdf:Group/xccdf:Rule", ns):
        total += 1
        if rule.get("id") == "Missing Rule":
            missing += 1
            version_text = rule.find("xccdf:version", ns).text
            missing_rules.append(version_text)
    implemented = total - missing
    stats = dict()
    stats["total"] = total
    stats["missing"] = missing
    stats["implemented"] = implemented
    if total != 0:
        stats["coverage"] = implemented / total * 100
    else:
        stats["coverage"] = 0
    stats["missing_rules"] = missing_rules
    return stats


def subtree_texts(el):
    if el is None:
        return ""
    else:
        return "".join(el.itertext())


def parse_description(rule_el, ns):
    description = subtree_texts(rule_el.find("./xccdf:description", ns))
    if description is None:
        return ""
    elif "<VulnDiscussion>" in description:
        _, _, after_open = description.partition("<VulnDiscussion>")
        before_close, _, _ = after_open.partition("</VulnDiscussion>")
        return before_close
    else:
        return description


def get_rules(root, ns):
    rules = []
    for group in root.findall(".//xccdf:Group", ns):
        rule = dict()
        rule_el = group.find("./xccdf:Rule", ns)
        rule["V-ID"] = group.get("id")
        ident = rule_el.find("./xccdf:ident", ns)
        rule["CCI"] = ident.text if ident.text is not None else ""
        rule["CAT"] = rule_el.get("severity")
        rule["title"] = rule_el.find("./xccdf:title", ns).text
        rule["SRG"] = group.find("./xccdf:title", ns).text
        rule["description"] = parse_description(rule_el, ns)
        check = rule_el.find("./xccdf:check/xccdf:check-content", ns)
        rule["check"] = check.text if check.text is not None else ""
        rule["fixtext"] = subtree_texts(rule_el.find("./xccdf:fixtext", ns))
        rule["version"] = rule_el.find("./xccdf:version", ns).text
        rule["mapped_rule"] = rule_el.get("id")
        rules.append(rule)
    return rules


def main():
    args = parse_args()
    tree = ET.parse(args.input)
    xccdf = determine_xccdf_tree_namespace(tree)
    ns = {"xccdf": xccdf}
    data = dict()
    root = tree.getroot()
    data["title"] = "Rules in " + root.find("./xccdf:title", ns).text
    data["stats"] = get_stats(root, ns)
    data["rules"] = get_rules(root, ns)
    render_template(data, STIG_TEMPLATE, args.output)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import argparse
import collections
import os
import re
import xml.etree.ElementTree as ET

import ssg.ansible
from ssg.constants import (
    ansible_system,
    bash_system,
    datastream_namespace,
    OSCAP_PROFILE,
    OSCAP_RULE,
    XCCDF12_NS,
)


DEFAULT_SELECTOR = "__DEFAULT"
HASH_ROW = "#" * 79
LANGUAGE_TO_SYSTEM = {"ansible": ansible_system, "bash": bash_system}
LANGUAGE_TO_TARGET = {"ansible": "playbook", "bash": "script"}
LANGUAGE_TO_EXTENSION = {"ansible": "yml", "bash": "sh"}
ANSIBLE_VAR_PATTERN = re.compile(
    "- name: XCCDF Value [^ ]+ # promote to variable\n  set_fact:\n"
    "    ([^:]+): (.+)\n  tags:\n    - always\n")


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-stream", required=True,
        help="Path of a SCAP source data stream file"
    )
    parser.add_argument(
        "--output-dir", required=True,
        help="Path of the output directory to save generated files"
    )
    parser.add_argument(
        "--product", required=True,
        help="Product ID, eg. 'rhel9'"
    )
    parser.add_argument(
        "--language", required=True, choices=["bash", "ansible"],
        help="Remediation language, either 'bash' or 'ansible'"
    )
    args = parser.parse_args()
    return args


def extract_ansible_vars(string):
    ansible_vars = {}
    for match in ANSIBLE_VAR_PATTERN.finditer(string):
        variable_name, value = match.groups()
        ansible_vars[variable_name] = value
    return ansible_vars


def indent(string, level):
    out = ""
    for line in string.splitlines():
        out += " " * level + line + "\n"
    return out


def extract_ansible_tasks(string):
    string = ANSIBLE_VAR_PATTERN.sub("\n", string)
    string = indent(string, 4)
    string = ssg.ansible.remove_trailing_whitespace(string)
    return [string]


def comment(text):
    commented_text = ""
    for line in text.split("\n"):
        trimmed_line = line.strip()
        if trimmed_line != "":
            commented_line = "# " + trimmed_line + "\n"
            commented_text += commented_line
    return commented_text


def get_selected_rules(profile):
    selected_rules = set()
    for select in profile.findall("./{%s}select" % XCCDF12_NS):
        if select.get("selected") == "true":
            id_ = select.get("idref")
            if id_.startswith(OSCAP_RULE):
                selected_rules.add(id_)
    return selected_rules


def get_value_refinenements(profile):
    refinements = {}
    for refine_value in profile.findall("./{%s}refine-value" % XCCDF12_NS):
        value_id = refine_value.get("idref")
        selector = refine_value.get("selector")
        refinements[value_id] = selector
    return refinements


def get_remediation(rule, system):
    for fix in rule.findall("./{%s}fix" % XCCDF12_NS):
        if fix.get("system") == system:
            return fix
    return None


def _get_all_itms(benchmark, element_name, callback):
    itms = {}
    el_xpath = ".//{%s}%s" % (XCCDF12_NS, element_name)
    for el in benchmark.findall(el_xpath):
        rule_id = el.get("id")
        itms[rule_id] = callback(el)
    return itms


def get_variable_values(variable):
    values = {}
    for value in variable.findall("./{%s}value" % (XCCDF12_NS)):
        selector = value.get("selector")
        if selector is None:
            selector = DEFAULT_SELECTOR
        if value.text is None:
            values["selector"] = ""
        else:
            values[selector] = value.text
    return values


def get_all_variables(benchmark):
    return _get_all_itms(benchmark, "Value", get_variable_values)


def expand_variables(fix_el, refinements, variables):
    content = fix_el.text[:]
    for sub_el in fix_el.findall("./{%s}sub" % XCCDF12_NS):
        variable_id = sub_el.get("idref")
        values = variables[variable_id]
        selector = refinements.get(variable_id, DEFAULT_SELECTOR)
        if selector == "default":
            return None
        if selector in values:
            value = values[selector]
        else:
            value = list(values.values())[0]
        content += value
        content += sub_el.tail
    return content


def ansible_vars_to_string(variables):
    var_strings = []
    indent = 4 * " "
    for k, v in variables.items():
        var_string = "%s%s: %s\n" % (indent, k, v)
        var_strings.append(var_string)
    return "".join(var_strings)


def ansible_tasks_to_string(tasks):
    # tasks need to be separated by 2 empty lines
    tasks_str = "\n\n".join(tasks)
    tasks_str = ssg.ansible.remove_too_many_blank_lines(tasks_str)
    return tasks_str


class ScriptGenerator:
    def __init__(self, language, product_id, ds_file_path, output_dir):
        self.language = language
        self.product_id = product_id
        self.ds_file_path = ds_file_path
        self.output_dir = output_dir
        self.get_benchmark_data()

    def generate_remediation_scripts(self):
        profile_xpath = "./{%s}component/{%s}Benchmark/{%s}Profile" % (
            datastream_namespace, XCCDF12_NS, XCCDF12_NS)
        for profile in self.ds.findall(profile_xpath):
            self.generate_profile_remediation_script(profile)

    def get_benchmark_data(self):
        self.ds = ET.parse(self.ds_file_path)
        self.ds_file_name = os.path.basename(self.ds_file_path)
        benchmark_xpath = "./{%s}component/{%s}Benchmark" % (
            datastream_namespace, XCCDF12_NS)
        benchmark = self.ds.find(benchmark_xpath)
        self.benchmark_id = benchmark.get("id")
        self.benchmark_version = benchmark.find(
            "./{%s}version" % XCCDF12_NS).text
        self.load_all_remediations(benchmark)
        self.variables = get_all_variables(benchmark)

    def load_all_remediations(self, benchmark):
        self.remediations = collections.OrderedDict()
        rule_xpath = ".//{%s}Rule" % (XCCDF12_NS)
        for rule_el in benchmark.findall(rule_xpath):
            rule_id = rule_el.get("id")
            system = LANGUAGE_TO_SYSTEM[self.language]
            remediation = get_remediation(rule_el, system)
            self.remediations[rule_id] = remediation

    def generate_profile_remediation_script(self, profile_el):
        if self.language == "ansible":
            output = self.create_output_ansible(profile_el)
        elif self.language == "bash":
            output = self.create_output_bash(profile_el)
        file_path = self.get_output_file_path(profile_el)
        with open(file_path, "wb") as f:
            f.write(output.encode("utf-8"))

    def get_output_file_path(self, profile_el):
        short_profile_id = profile_el.get("id").replace(OSCAP_PROFILE, "")
        file_name = "%s-%s-%s.%s" % (
            self.product_id, LANGUAGE_TO_TARGET[self.language],
            short_profile_id, LANGUAGE_TO_EXTENSION[self.language])
        file_path = os.path.join(self.output_dir, file_name)
        return file_path

    def create_output_ansible(self, profile_el):
        ansible_vars, ansible_tasks = self.collect_ansible_vars_and_tasks(
            profile_el)
        header = self.create_header(profile_el)
        profile_id = profile_el.get("id")
        vars_str = ansible_vars_to_string(ansible_vars)
        tasks_str = ansible_tasks_to_string(ansible_tasks)
        output = (
            "%s\n"
            "- name: Ansible Playbook for %s\n"
            "  hosts: all\n"
            "  vars:\n"
            "%s"
            "  tasks:\n"
            "%s"
            % (header, profile_id, vars_str, tasks_str))
        return output

    def collect_ansible_vars_and_tasks(self, profile_el):
        selected_rules = get_selected_rules(profile_el)
        refinements = get_value_refinenements(profile_el)
        all_vars = collections.OrderedDict()
        all_tasks = []
        for rule_id, fix_el in self.remediations.items():
            if rule_id not in selected_rules:
                continue
            rule_vars, rule_tasks = self.generate_ansible_rule_remediation(
                fix_el, refinements)
            all_vars.update(rule_vars)
            all_tasks.extend(rule_tasks)
        return (all_vars, all_tasks)

    def create_output_bash(self, profile):
        output = []
        selected_rules = get_selected_rules(profile)
        refinements = get_value_refinenements(profile)
        header = self.create_header(profile)
        output.append(header)
        total = len(selected_rules)
        current = 1
        for rule_id in self.remediations:
            if rule_id not in selected_rules:
                continue
            status = (current, total)
            rule_remediation = self.generate_bash_rule_remediation(
                rule_id, status, refinements)
            output.append(rule_remediation)
            current += 1
        return "".join(output)

    def create_header(self, profile):
        if self.language == "ansible":
            shebang_with_newline = "---\n"
            remediation_type = "Ansible Playbook"
            how_to_apply = (
                "# $ ansible-playbook -i \"localhost,\""
                " -c local playbook.yml\n"
                "# $ ansible-playbook -i \"192.168.1.155,\" playbook.yml\n"
                "# $ ansible-playbook -i inventory.ini playbook.yml\n"
            )
        elif self.language == "bash":
            shebang_with_newline = "#!/usr/bin/env bash\n"
            remediation_type = "Bash Remediation Script"
            how_to_apply = "# $ sudo ./remediation-script.sh\n"
        profile_title = profile.find("./{%s}title" % XCCDF12_NS).text
        description = profile.find("./{%s}description" % XCCDF12_NS).text
        commented_profile_description = comment(description)
        xccdf_version_name = "1.2"
        profile_id = profile.get("id")
        fix_header = (
            "%s"
            "%s\n"
            "#\n"
            "# %s for %s\n"
            "#\n"
            "# Profile Description:\n"
            "%s"
            "#\n"
            "# Profile ID:  %s\n"
            "# Benchmark ID:  %s\n"
            "# Benchmark Version:  %s\n"
            "# XCCDF Version:  %s\n"
            "#\n"
            "# This file can be generated by OpenSCAP using:\n"
            "# $ oscap xccdf generate fix --profile %s --fix-type %s %s\n"
            "#\n"
            "# This %s is generated from an XCCDF profile without"
            " preliminary evaluation.\n"
            "# It attempts to fix every selected rule, even if the system is"
            " already compliant.\n"
            "#\n"
            "# How to apply this %s:\n"
            "%s"
            "#\n"
            "%s\n\n" % (
                shebang_with_newline, HASH_ROW, remediation_type,
                profile_title, commented_profile_description, profile_id,
                self.benchmark_id, self.benchmark_version, xccdf_version_name,
                profile_id, self.language, self.ds_file_name,
                remediation_type, remediation_type, how_to_apply, HASH_ROW))
        return fix_header

    def generate_bash_rule_remediation(self, rule_id, status, refinements):
        current, total = status
        fix_el = self.remediations[rule_id]
        output = []
        header = (
            "%s\n"
            "# BEGIN fix (%s / %s) for '%s'\n"
            "%s\n" % (HASH_ROW, current, total, rule_id, HASH_ROW))
        output.append(header)
        begin_msg = "(>&2 echo \"Remediating rule %s/%s: '%s'\")\n" % (
            current, total, rule_id)
        output.append(begin_msg)
        expanded_remediation = None
        if fix_el is not None:
            expanded_remediation = expand_variables(
                fix_el, refinements, self.variables)
        if expanded_remediation is not None:
            output.append(expanded_remediation)
        else:
            warning = (
                "(>&2 echo \"FIX FOR THIS RULE '%s' IS MISSING!\")\n" %
                (rule_id))
            output.append(warning)
        end_msg = "\n# END fix for '%s'\n\n" % (rule_id)
        output.append(end_msg)
        return "".join(output)

    def generate_ansible_rule_remediation(self, fix_el, refinements):
        rule_vars = {}
        tasks = []
        expanded_remediation = None
        if fix_el is not None:
            expanded_remediation = expand_variables(
                fix_el, refinements, self.variables)
        if expanded_remediation is not None:
            rule_vars = extract_ansible_vars(expanded_remediation)
            tasks = extract_ansible_tasks(expanded_remediation)
        return rule_vars, tasks


def main():
    args = parse_args()
    if not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
    sg = ScriptGenerator(
        args.language, args.product, args.data_stream, args.output_dir)
    sg.generate_remediation_scripts()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
import argparse
import json
import pathlib
import sys
import xml.etree.ElementTree as ET

from ssg.constants import OSCAP_RULE, XCCDF12_NS

SSG_ROOT = pathlib.Path(__file__).resolve().parent.parent
BUILD_DIR = SSG_ROOT.joinpath("build")
RULE_DIR_JSON = BUILD_DIR.joinpath("rule_dirs.json")
EPILOG = """
This script lists rules that are not used in any data streams.
It requires that all products (and derivatives) are built.
To do this run ./build_product --derivatives
The script has the following return codes:
    0 - All rules are used in the data streams,
    1 - Some rules are not used in the data streams,
    2 - Not all products are built, and
    3 - rule_dirs.json does not exist.
"""


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="List rules that are not used in any "
                                                 "data streams."
                                                 "Note that script requires that all products "
                                                 "(and derivatives) are built.", epilog=EPILOG)
    parser.add_argument("--root",
                        help="Root directory of the SSG git repository", default=SSG_ROOT)
    parser.add_argument("--json", help="Path to rule_dir.json file",
                        default=RULE_DIR_JSON.absolute())
    parser.add_argument("--force",
                        help="Force the operation even if all products are not built",
                        action="store_true")
    return parser.parse_args()


def _get_ds_rules(datastream_files):
    ds_rules = set()
    for ds in datastream_files:
        root = ET.parse(ds).getroot()
        root_elements = list(root.findall(".//{%s}Rule" % XCCDF12_NS))
        for rule in root_elements:
            rule_id = rule.get("id").removeprefix(OSCAP_RULE)
            ds_rules.add(rule_id)
    return ds_rules


def _get_product_count(products_path):
    products_count = 0
    for product in products_path.iterdir():
        if product.is_dir() and product.name != "example":
            products_count += 1
    return products_count


def main() -> int:
    args = _parse_args()
    root_path = pathlib.Path(args.root)
    products_path = root_path.joinpath("products")
    build_dir = root_path.joinpath("build")
    rule_dir_path = pathlib.Path(args.json)
    if not rule_dir_path.exists():
        print(f"Rule directory {rule_dir_path} does not exist.", file=sys.stderr)
        print("Hint run: ./utils/rule_dir_json.py", file=sys.stderr)
        return 3
    rule_dir_json = json.loads(rule_dir_path.read_text())
    all_rules = set(rule_dir_json.keys())
    products_count = _get_product_count(products_path)
    datastream_files = list(build_dir.glob("ssg-*-ds.xml"))
    ds_products = set()
    for ds in datastream_files:
        ds_products.add(ds.name.split("-")[1])
    if products_count > len(datastream_files):
        print("Not all products are built, cowardly refusing to continue.", file=sys.stderr)
        print(f"Products: {products_count}, data streams: {len(datastream_files)}", file=sys.stderr)
        print("Hint: run ./build_product --derivatives", file=sys.stderr)
        return 2
    ds_rules = _get_ds_rules(datastream_files)
    disuse_rules = all_rules - ds_rules
    if not disuse_rules:
        print("All rules are used in the datastream files.")
        return 0
    print("The following rules are not used in ANY of the provided data stream files:")
    print("\n".join(disuse_rules))
    return 1


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/python3

import collections
import sys
import pathlib
import re

import pygments
import pygments.lexers
import pygments.formatters

import ssg.build_yaml
import ssg.controls
import ssg.yaml
import ssg.jinja
import template_renderer
from rendering import common


POLICY_FIELDS_MAPPING = dict(
    checktext="checks",
    fixtext="fixes",
    srg_requirement="prose",
    vuldiscussion="prose",
)


class Namespace:
    def __init__(self):
        self.policy_specific_content = collections.defaultdict(dict)

    def contains_something(self):
        copy_of_attributes = dict(vars(self))

        if self.policy_specific_content:
            return True
        else:
            copy_of_attributes.pop("policy_specific_content")

        for stuff in copy_of_attributes:
            if stuff:
                return True
        return False

    def sub_newlines_with_br(self):
        attributes = vars(self)
        for attr, value in attributes.items():
            if not value:
                continue
            attributes[attr] = re.sub(r"\n+", r"<br/>\n", value)


class HtmlOutput(template_renderer.Renderer):
    TEMPLATE_NAME = "rendering/rule-template.html"

    def process_rule(self, rule_id):
        self.built_content_path = pathlib.Path(self.built_content_path)

        rule_fname = self.built_content_path / "rules" / (rule_id + ".yml")
        try:
            rule = ssg.build_yaml.Rule.from_yaml(rule_fname, self.env_yaml)
        except RuntimeError as exc:
            msg = (
                f"{str(exc)}. Make sure that the product is compiled, "
                f"and '{rule_id}' is a short rule ID.")
            print(msg, file=sys.stderr)
            sys.exit(1)

        prose = Namespace()
        prose.description = common.resolve_var_substitutions(rule.description)
        prose.rationale = common.resolve_var_substitutions(rule.rationale)
        prose.sub_newlines_with_br()

        self.template_data["prose"] = prose
        self.template_data["fixes"] = self._get_fixes(rule_id)
        self.template_data["checks"] = self._get_checks(rule_id)
        self.template_data["rule"] = rule

        self._add_policy_content_to_categories(rule.policy_specific_content)

    def _add_policy_content_to_categories(self, policy_specific_content):
        for policy, contents in policy_specific_content.items():
            for field_name, text in contents.items():
                target_category = POLICY_FIELDS_MAPPING.get(field_name)
                if target_category is None:
                    continue

                target_destination = self.template_data[target_category]
                text = re.sub(r"\n+", r"<br/>\n", text)
                target_destination.policy_specific_content[policy][field_name] = text

    def _highlight_file(self, fname, lexer):
        with open(fname, "r") as f:
            code = f.read()
            code = common.resolve_var_substitutions(code)
            return pygments.highlight(code, lexer, pygments.formatters.HtmlFormatter())

    def _get_fixes(self, rule_id):
        fixes = Namespace()

        basename = rule_id + ".sh"
        bash_loc = self.built_content_path / "fixes" / "bash" / basename
        if not bash_loc.exists():
            bash_loc = self.built_content_path / "fixes_from_templates" / "bash" / basename
        if bash_loc.exists():
            fixes.bash = self._highlight_file(bash_loc, pygments.lexers.BashLexer())

        basename = rule_id + ".yml"
        ansible_loc = self.built_content_path / "fixes" / "ansible" / basename
        if not ansible_loc.exists():
            ansible_loc = self.built_content_path / "fixes_from_templates" / "ansible" / basename
        if ansible_loc.exists():
            fixes.ansible = self._highlight_file(ansible_loc, pygments.lexers.YamlLexer())

        return fixes

    def _get_checks(self, rule_id):
        checks = Namespace()

        basename = rule_id + ".xml"
        oval_loc = self.built_content_path / "checks" / "oval" / basename
        if not oval_loc.exists():
            oval_loc = self.built_content_path / "checks_from_templates" / "oval" / basename
        if oval_loc.exists():
            checks.oval = self._highlight_file(oval_loc, pygments.lexers.XmlLexer())

        return checks


def parse_args():
    parser = HtmlOutput.create_parser(
        "Pass a short rule ID, and the script will render the rule fields and "
        "associated content in a form of a single HTML file, "
        "so consistency of the rule can be conveniently examined. "
        "To get syntax highlighting, you will need a highlight.css file in the same directory, "
        "you can generate one using pygmentize e.g. by running "
        "pygmentize -S solarized-dark -f html > highlight.css")
    parser.add_argument(
        "rule", metavar="RULE_ID", help="The rule short ID")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    renderer = HtmlOutput(args.product, args.build_dir)
    renderer.process_rule(args.rule)
    renderer.output_results(args)



#!/usr/bin/python3

import argparse
import collections
import os

import ssg.build_stig
import ssg.build_yaml
import ssg.constants
import ssg.jinja
from utils.template_renderer import render_template


TABLE_DIR = os.path.join(os.path.dirname(__file__), "tables")
SRGMAP_TEMPLATE = os.path.join(TABLE_DIR, "srgmap_template.html")
SRGMAP_FLAT_TEMPLATE = os.path.join(TABLE_DIR, "srgmap_flat_template.html")


def parse_args():
    parser = argparse.ArgumentParser(
        description="Create a SRG map and a SRG flat map, both at once")
    parser.add_argument(
        "--build-dir", default="build", help="Path to the build directory")
    parser.add_argument(
        "product", help="Short product ID, eg. rhel8")
    parser.add_argument(
        "srgs", help="SRG file, eg. /shared/references/disa-os-srg-v2r1.xml")
    parser.add_argument("srgmap", help="Output SRG map path")
    parser.add_argument("srgmap_flat", help="Output flat SRG map path")
    return parser.parse_args()


def get_rules_by_srgid(build_dir, product):
    profile_filename = os.path.join(
        build_dir, product, "profiles", "stig.profile")
    profile = ssg.build_yaml.Profile.from_yaml(profile_filename)
    rules_root = os.path.join(build_dir, product, "rules")
    rules = collections.defaultdict(list)
    for rule_id in profile.selected:
        rule_filename = os.path.join(rules_root, rule_id + ".yml")
        rule = ssg.build_yaml.Rule.from_yaml(rule_filename)
        if "srg" in rule.references:
            for srgid in rule.references["srg"]:
                rules[srgid].append(rule)
    return rules


if __name__ == "__main__":
    args = parse_args()
    data = dict()
    data["srgs"] = ssg.build_stig.parse_srgs(args.srgs)
    data["rules_by_srgid"] = get_rules_by_srgid(args.build_dir, args.product)
    data["full_name"] = ssg.utils.product_to_name(args.product)
    render_template(data, SRGMAP_TEMPLATE, args.srgmap)
    render_template(data, SRGMAP_FLAT_TEMPLATE, args.srgmap_flat)



#!/usr/bin/python3

from __future__ import print_function

import os
import sys

import ssg.build_yaml

import tables.table_renderer


class HtmlOutput(tables.table_renderer.TableHtmlOutput):
    TEMPLATE_NAME = "tables/profile_tables_template.html"

    def __init__(self, * args, ** kwargs):
        profile_id = kwargs.pop("profile_id")
        super(HtmlOutput, self).__init__(* args, ** kwargs)

        profile_path = os.path.join(self.built_content_path, "profiles", (profile_id + ".profile"))
        self.profile = ssg.build_yaml.Profile.from_yaml(profile_path, self.env_yaml)
        self.profile_variables = self.profile.variables

    def _get_var_value(self, varname):
        try:
            return self.profile_variables[varname]
        except KeyError:
            return super(HtmlOutput, self)._get_var_value(varname)

    def _get_eligible_rules(self, refcat):
        eligible_rule_ids = self.profile.selected
        filenames = [os.path.join(self.rules_root, rid + ".yml") for rid in eligible_rule_ids]
        return [ssg.build_yaml.Rule.from_yaml(f, self.env_yaml) for f in filenames]

    def _generate_shortened_ref(self, reference, rule):
        shortened_ref = super(HtmlOutput, self)._generate_shortened_ref(reference, rule)
        if shortened_ref == self.DEFAULT_SHORTENED_REF and self.verbose:
            print("Rule '{rid}' is included in the profile {profile_id}, "
                  "but doesn't contain the {ref_id} reference"
                  .format(rid=rule.id_, profile_id=self.profile.id_, ref_id=reference.id),
                  file=sys.stderr)
        return shortened_ref

    def process_rules(self, reference):
        super(HtmlOutput, self).process_rules(reference)

        self.template_data["profile"] = self.profile


def update_parser(parser):
    parser.add_argument(
        "profile", metavar="PROFILE_ID", help="The ID of the profile")


def parse_args():
    parser = HtmlOutput.create_parser(
        "Generate HTML table that maps references to rules in context of a profile "
        "using compiled rules and compiled profiles as source of data.")
    tables.table_renderer.update_parser(parser)
    update_parser(parser)
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    renderer = HtmlOutput(
        args.product, args.build_dir, profile_id=args.profile, verbose=args.verbose)
    reference = ssg.constants.REFERENCES[args.refcategory]
    renderer.process_rules(reference)
    renderer.output_results(args)



from ssg.utils import parse_template_boolean_value


def preprocess(data, lang):
    data["check_root_user"] = parse_template_boolean_value(data, parameter="check_root_user", default_value=False)
    if lang == "bash":
        if "syscall_grouping" in data:
            # Make it easier to tranform the syscall_grouping into a Bash array
            data["syscall_grouping"] = " ".join(data["syscall_grouping"])
    elif lang == "ansible":
        if "attr" in data:
            # Tranform the syscall into a Ansible list
            data["attr"] = [data["attr"]]
        if "syscall_grouping" not in data:
            # Ensure that syscall_grouping is a list
            data["syscall_grouping"] = []

    return data



#!/usr/bin/python3

from __future__ import print_function

"""
Takes given XCCDF or data stream and for every profile in it it generates one
OpenSCAP HTML guide. Also generates an index file that lists all the profiles
and allows the user to navigate between them.

Author: Martin Preisler <mpreisle@redhat.com>
"""

import os.path
import argparse
import threading
import sys

import ssg.build_guides
import ssg.xccdf
import ssg.utils
import ssg.xml


def parse_args():
    p = argparse.ArgumentParser()

    sp = p.add_subparsers(help="actions")

    make_sp = sp.add_parser("build", help="Build all the HTML guides")
    make_sp.set_defaults(cmd="build")

    input_sp = sp.add_parser("list-inputs", help="Generate input list")
    input_sp.set_defaults(cmd="list_inputs")

    output_sp = sp.add_parser("list-outputs", help="Generate output list")
    output_sp.set_defaults(cmd="list_outputs")

    p.add_argument("-j", "--jobs", type=int, action="store",
                   default=ssg.utils.get_cpu_count(),
                   help="how many jobs should be processed in parallel")
    p.add_argument("-i", "--input", action="store", required=True,
                   help="input file, can be XCCDF or Source data stream")
    p.add_argument("-o", "--output", action="store", required=True,
                   help="output directory")

    return p.parse_args()


def main():
    args = parse_args()

    input_path, input_basename, path_base, output_dir = \
        ssg.build_guides.get_path_args(args)
    index_path = os.path.join(output_dir, "%s-guide-index.html" % (path_base))

    if args.cmd == "list_inputs":
        print(input_path)
        sys.exit(0)

    input_tree = ssg.xml.ElementTree.parse(input_path)
    benchmarks = ssg.xccdf.get_benchmark_id_title_map(input_tree)
    if len(benchmarks) == 0:
        raise RuntimeError(
            "Expected input file '%s' to contain at least 1 xccdf:Benchmark. "
            "No Benchmarks were found!" %
            (input_path)
        )

    benchmark_profile_pairs = ssg.build_guides.get_benchmark_profile_pairs(
        input_tree, benchmarks)

    if args.cmd == "list_outputs":
        guide_paths = ssg.build_guides.get_output_guide_paths(benchmarks,
                                                              benchmark_profile_pairs,
                                                              path_base, output_dir)

        for guide_path in guide_paths:
            print(guide_path)
        print(index_path)
        sys.exit(0)

    index_links, index_options, index_initial_src, queue = \
        ssg.build_guides.fill_queue(benchmarks, benchmark_profile_pairs,
                                    input_path, path_base, output_dir)

    workers = []
    for worker_id in range(args.jobs):
        worker = threading.Thread(
            name="Guide generate worker #%i" % (worker_id),
            target=lambda queue=queue: ssg.build_guides.builder(queue)
        )
        workers.append(worker)
        worker.daemon = True
        worker.start()

    for worker in workers:
        worker.join()

    if queue.unfinished_tasks > 0:
        raise RuntimeError("Some of the guides were not exported successfully")

    index_source = ssg.build_guides.build_index(benchmarks, input_basename,
                                                index_links, index_options,
                                                index_initial_src)

    with open(index_path, "wb") as f:
        f.write(index_source.encode("utf-8"))


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import argparse
import os

import ssg.build_ovals


def main():
    parser = argparse.ArgumentParser(
        description="Convert shorthand OVAL file to a valid full OVAL.")
    parser.add_argument("input", help="Input shorthand OVAL file")
    parser.add_argument("output", help="Output OVAL file")
    args = parser.parse_args()
    env_yaml = {"rule_id": os.path.basename(args.input)}
    ssg.build_ovals.expand_shorthand(args.input, args.output, env_yaml)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

from __future__ import print_function

import os
import re
import shutil
import subprocess
import sys
import tempfile

import ssg.build_ovals
import ssg.constants
import ssg.jinja
import ssg.oval
import ssg.rules
import ssg.utils
import ssg.yaml
import ssg.build_yaml
import ssg.rule_yaml
import ssg.id_translate


class OVALTester(object):
    def __init__(self, verbose):
        self.mock_env_yaml = {"rule_id": "test"}
        self.result = True
        self.verbose = verbose

    def _get_result(self, oscap_output):
        pattern = re.compile(
            r"^Definition oval:[A-Za-z0-9_\-\.]+:def:[1-9][0-9]*: (\w+)$")
        for line in oscap_output.splitlines():
            matched = pattern.match(line)
            if matched:
                return matched.group(1)
        return None

    def _create_config_file(self, config_file_content, tmp_dir):
        config_file_path = os.path.join(tmp_dir, "config")
        if config_file_content:
            with open(config_file_path, "w") as f:
                f.write(config_file_content)
        return config_file_path

    def _create_oval(self, oval_content, config_file_path, tmp_dir):
        oval_content = oval_content.replace("CONFIG_FILE", config_file_path)
        shorthand_path = os.path.join(tmp_dir, "shorthand")
        with open(shorthand_path, "w") as f:
            f.write(oval_content)
        oval_path = os.path.join(tmp_dir, "oval.xml")
        ssg.build_ovals.expand_shorthand(
            shorthand_path, oval_path, self.mock_env_yaml)
        return oval_path

    def _evaluate_oval(self, oval_path):
        results_path = oval_path + ".results.xml"
        oscap_command = [
            "oscap", "oval", "eval", "--results", results_path, oval_path]
        oscap_process = subprocess.Popen(
            oscap_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        oscap_stdout, oscap_stderr = oscap_process.communicate()
        def_result = self._get_result(oscap_stdout.decode("utf-8"))
        return def_result, results_path

    def test(self, description, oval_content, config_file_content,
             expected_result):
        """
        Execute a test.
        description: a very short description to be displayed in test output
        oval_content: content of the OVAL shorthand file written in a way you
                      write OVALs in SSG rules (not a valid OVAL)
        config_file_content: content of the text configuration file that the
                             OVAL will check
        expected_result: expected result of evaluation of the OVAL definition
        """
        try:
            tmp_dir = tempfile.mkdtemp()
            config_file_path = self._create_config_file(
                config_file_content, tmp_dir)
            oval_path = self._create_oval(
                oval_content, config_file_path, tmp_dir)
            result, results_path = self._evaluate_oval(oval_path)
            msg = ("OVAL Definition was evaluated as %s, but expected result "
                   "is %s.") % (result, expected_result)
            assert result == expected_result, msg
            if self.verbose:
                print("Test: %s: PASS" % (description))
            self.result = self.result and True
        except AssertionError as e:
            if self.verbose:
                print("Test: %s: FAIL" % (description))
                print("    " + str(e))
            self.result = False
        finally:
            shutil.rmtree(tmp_dir)

    def finish(self):
        """
        Exit test with an appropriate return code.
        """
        sys.exit(0 if self.result else 1)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('../tests'))

# since these files live outside of root doc directory, we need to create a symlink to a dir under where conf.py lives
# so myst-parser can properly render them
filepaths = ["tests/README.md"]
for filepath in filepaths:
    try:
        path, _ = filepath.split("/")
        os.mkdir(os.path.join(path))
        os.symlink(os.path.join(os.path.abspath('..'), filepath), os.path.join(filepath))
    except FileExistsError as e:
        # ignore if file exists
        print("Warning: {}".format(e))

# -- Project information -----------------------------------------------------

project = 'ComplianceAsCode/content'
copyright = 'ComplianceAsCode'
author = 'ComplianceAsCode'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['myst_parser',
    'sphinxcontrib.apidoc',
    'sphinx_rtd_theme',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinxcontrib.jinjadomain',
    'sphinxcontrib.autojinja.jinja',
]

apidoc_module_dir = '../ssg'
apidoc_output_dir = 'api'
apidoc_excluded_paths = ['tests', 'build-scripts']
apidoc_separate_modules = True

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Required for autojinja to work
jinja_template_path = os.path.abspath('..')

source_suffix = ['.rst', '.md']

# Allow linking to (most) headers
myst_heading_anchors = 3



import os
import json

from ssg.build_profile import XCCDFBenchmark
from ssg.utils import mkdir_p


OFF_JSON_TO_HTML = False

try:
    from json2html import json2html
except ImportError:
    OFF_JSON_TO_HTML = True


def _process_stats_content(profile, bash_fixes_count, content, content_filepath):
    link = """<a href="{}"><div style="height:100%;width:100%">{}</div></a>"""
    count = len(profile[content])
    if content == "ansible_parity":
        # custom text link for ansible parity
        count = link.format(
            content_filepath,
            "{} out of {} ({}%)".format(
                bash_fixes_count - count,
                bash_fixes_count,
                int(((bash_fixes_count - count) / bash_fixes_count) * 100),
            ),
        )
    count_href_element = link.format(content_filepath, count)
    profile["{}_count".format(content)] = count_href_element


def _filter_profile_for_html_stats(profile, filtered_output, content_path):
    content_list = [
        "rules",
        "missing_stigid_refs",
        "missing_stigref_refs",
        "missing_ccn_refs",
        "missing_cis_refs",
        "missing_hipaa_refs",
        "missing_anssi_refs",
        "missing_ospp_refs",
        "missing_pcidss4_refs",
        "missing_cui_refs",
        "missing_ovals",
        "missing_sces",
        "missing_bash_fixes",
        "missing_ansible_fixes",
        "missing_ignition_fixes",
        "missing_kubernetes_fixes",
        "missing_puppet_fixes",
        "missing_anaconda_fixes",
        "missing_cces",
        "ansible_parity",
        "implemented_checks",
        "implemented_fixes",
        "missing_checks",
        "missing_fixes",
    ]

    bash_fixes_count = profile["rules_count"] - profile["missing_bash_fixes_count"]

    for content in content_list:
        content_file = "{}_{}.txt".format(profile["profile_id"], content)
        content_filepath = os.path.join("content", content_file)
        count = len(profile[content])
        if count > 0:
            _process_stats_content(profile, bash_fixes_count, content, content_filepath)
            with open(os.path.join(content_path, content_file), "w+") as f:
                f.write("\n".join(profile[content]))
        else:
            profile["{}_count".format(content)] = count
        del profile[content]
    filtered_output.append(profile)


def _get_profiles(args):
    benchmark = XCCDFBenchmark(args.benchmark, args.product)
    ret = []
    if args.profile:
        ret.append(benchmark.show_profile_stats(args.profile, args))
    else:
        ret.extend(benchmark.show_all_profile_stats(args))
    return ret


def _generate_html_stats(args, profiles):
    filtered_output = []
    output_path = "./"
    if args.output:
        output_path = args.output
        mkdir_p(output_path)

    content_path = os.path.join(output_path, "content")
    mkdir_p(content_path)

    for profile in profiles:
        _filter_profile_for_html_stats(profile, filtered_output, content_path)

    with open(os.path.join(output_path, "statistics.html"), "w+") as f:
        f.write(json2html.convert(json=json.dumps(filtered_output), escape=False))


def command_stats(args):
    profiles = _get_profiles(args)

    if args.format == "json":
        print(json.dumps(profiles, indent=4))

    elif args.format == "html":
        if OFF_JSON_TO_HTML:
            print("No module named 'json2html'. Please install module to enable this function.")
            return
        _generate_html_stats(args, profiles)

    elif args.format == "csv":
        # we can assume ret has at least one element
        # CSV header
        print(",".join(profiles[0].keys()))
        for line in profiles:
            print(",".join([str(value) for value in line.values()]))



"""
Common functions for processing Fixes in SSG
"""

from __future__ import absolute_import
from __future__ import print_function

import os

from .build_remediations import parse_from_file_without_jinja
from .utils import read_file_list, parse_platform

from .build_remediations import REMEDIATION_TO_EXT_MAP as REMEDIATION_MAP


def get_fix_path(rule_obj, lang, fix_id):
    """
    Return the full path to the fix for the given language and fix_id in the rule described by rule_obj.

    Args:
        rule_obj (dict): A dictionary containing information about the rule, including dir', 'id',
                         and 'remediations'.
        lang (str): The language for which the fix is required.
        fix_id (str): The identifier for the specific fix.

    Returns:
        str: The full path to the fix file.

    Raises:
        ValueError: If the rule_obj is malformed or if the fix_id is unknown for the given rule_id
                    and language.
    """
    if not fix_id.endswith(REMEDIATION_MAP[lang]):
        fix_id += REMEDIATION_MAP[lang]

    if ('dir' not in rule_obj or 'id' not in rule_obj or
        'remediations' not in rule_obj or lang not in rule_obj['remediations']):
        raise ValueError("Malformed rule_obj")

    if fix_id not in rule_obj['remediations'][lang]:
        raise ValueError("Unknown fix_id:%s for rule_id:%s and lang:%s" %
                         (fix_id, rule_obj['id'], lang))

    return os.path.join(rule_obj['dir'], lang, fix_id)


def get_fix_contents(rule_obj, lang, fix_id):
    """
    Retrieve the path and contents of a specific fix.

    Args:
        rule_obj (object): The rule object containing the fix information.
        lang (str): The language of the fix.
        fix_id (str): The identifier of the fix.

    Returns:
        tuple: A tuple containing the path to the fix and the contents of the fix.
    """
    fix_path = get_fix_path(rule_obj, lang, fix_id)
    fix_contents = read_file_list(fix_path)

    return fix_path, fix_contents


def applicable_platforms(fix_path):
    """
    Determines the applicable platforms for a given fix.

    Args:
        fix_path (str): The file path to the fix configuration file.

    Returns:
        list: A list of platforms that the fix applies to.

    Raises:
        ValueError: If the fix configuration is malformed or missing the 'platform' key.
    """
    _, config = parse_from_file_without_jinja(fix_path)

    if 'platform' not in config:
        raise ValueError("Malformed fix: missing platform" % fix_path)

    return parse_platform(config['platform'])


def find_platform_line(fix_contents):
    """
    Parses the fix content to determine the line number that the platforms configuration option is on.

    Note:
        If the configuration specification changes, please update the corresponding parsing in
        ssg.build_remediations.parse_from_file_with_jinja(...).

    Args:
        fix_contents (list of str): The contents of the configuration file as a list of strings.

    Returns:
        int or None: The line number of the platform configuration option, or None if not found.
    """
    matched_line = None
    for line_num in range(0, len(fix_contents)):
        line = fix_contents[line_num]
        if line.startswith('#') and '=' in line:
            key, value = line.strip('#').split('=', 1)
            if key.strip() == 'platform':
                matched_line = line_num

    return matched_line

def set_applicable_platforms(fix_contents, new_platforms):
    """
    Modifies the given fix contents to update the platforms to the new platforms.

    Args:
        fix_contents (list of str): The contents of the fix file as a list of strings.
        new_platforms (list of str): A list of new platforms to set in the fix file.

    Returns:
        list of str: The modified contents with the updated platforms.

    Raises:
        ValueError: If the platform line cannot be found in the fix contents.
    """
    platform_line = find_platform_line(fix_contents)
    if platform_line is None:
        raise ValueError("When parsing config file, unable to find platform "
                         "line!\n\n%s" % "\n".join(fix_contents))

    new_platforms_str = "# platform = " + ",".join(sorted(new_platforms))

    new_contents = fix_contents[0:platform_line]
    new_contents.extend([new_platforms_str])
    new_contents.extend(fix_contents[platform_line+1:])

    return new_contents



#!/usr/bin/env python3

import pathlib
import pandas as pd

import utils.srg_export.data

HTML_OUTPUT_TEMPLATE = '''
<html>
<head>
<title>{title}</title>
<style type="text/css">
table
{{
    border-collapse:collapse;
}}
table, th, td
{{
    border: 2px solid #dcdcdc;
    border-left: none;
    border-right: none;
    vertical-align: top;
    padding: 2px;
    font-family: verdana,arial,sans-serif;
    font-size:11px;
}}
pre {{
    white-space: pre-wrap;
    white-space: -moz-pre-wrap !important;
    word-wrap:break-word;
}}
table tr:nth-child(2n+2) {{ background-color: #f4f4f4; }}
thead
{{
    display: table-header-group;
    font-weight: bold;
    background-color: #dedede;
}}
</style>
</head>
<body>
    {table}
</body>
</html>
'''


def handle_dict(data: list, output_path: str, title: str) -> None:
    """
    Given a dict with the fields for the srg export, create an HTML file
    """
    pd.set_option('colheader_justify', 'center')
    df = pd.DataFrame(data, index=None)
    df.fillna("", inplace=True)
    df = df.reindex(utils.srg_export.data.HEADERS, axis=1)
    with open(pathlib.Path(output_path), 'w+') as f:
        html_table = df.to_html().replace("\\n", "<br>")
        html_page = HTML_OUTPUT_TEMPLATE.format(title=title, table=html_table)
        f.write(html_page)



"""
Common functions for enabling derivative products
"""

from __future__ import absolute_import
from __future__ import print_function

import re
from .xml import ElementTree
from .constants import (
    standard_profiles,
    OSCAP_VENDOR,
    cpe_dictionary_namespace,
    oval_namespace,
    datastream_namespace,
)
from .build_cpe import ProductCPEs, get_linked_cpe_oval_document
from .products import load_product_yaml


def add_cpes(elem, namespace, mapping):
    """
    Adds derivative CPEs next to RHEL ones, checks XCCDF elements of given
    namespace.
    """

    affected = False

    for child in list(elem):
        affected = affected or add_cpes(child, namespace, mapping)

    # precompute this so that we can affect the tree while iterating
    children = list(elem.findall(".//{%s}platform" % (namespace)))

    for child in children:
        idref = child.get("idref")
        if idref in mapping:
            new_platform = ElementTree.Element("{%s}platform" % (namespace))
            new_platform.set("idref", mapping[idref])
            # this is done for the newline and indentation
            new_platform.tail = child.tail

            index = list(elem).index(child)
            # insert it right after the respective RHEL CPE
            elem.insert(index + 1, new_platform)

            affected = True

    return affected


def get_cpe_item(product_yaml, cpe_ref, cpe_items_dir):
    product_cpes = ProductCPEs()
    product_cpes.load_cpes_from_directory_tree(cpe_items_dir, product_yaml)
    return product_cpes.get_cpe(cpe_ref)


def add_cpe_item_to_dictionary(
    tree_root, product_yaml_path, cpe_ref, id_name, cpe_items_dir
):
    cpe_list = tree_root.find(".//{%s}cpe-list" % cpe_dictionary_namespace)
    if cpe_list is not None:
        product_yaml = load_product_yaml(product_yaml_path)
        cpe_item = get_cpe_item(product_yaml, cpe_ref, cpe_items_dir)
        cpe_item.content_id = id_name
        cpe_item.set_cpe_oval_def_id()
        cpe_list.append(
            cpe_item.to_xml_element("ssg-%s-cpe-oval.xml" % product_yaml.get("product"))
        )
        return cpe_item.cpe_oval_short_def_id
    return None


def add_element_to(oval_root, tag_name, component_element):
    xml_el = oval_root.find(".//{%s}%s" % (oval_namespace, tag_name))
    if xml_el is None:
        xml_el = ElementTree.Element("{%s}%s" % (oval_namespace, tag_name))
        oval_root.append(xml_el)
    if xml_el.find("%s[@id='%s']" % (component_element.tag, component_element.get("id"))) is None:
        xml_el.append(component_element)


def add_oval_components_to_oval_xml(oval_root, tag_name, component_dict):
    for component in component_dict.values():
        add_element_to(oval_root, tag_name, component.get_xml_element())


def get_cpe_oval_root(root):
    for component_el in root.findall("./{%s}component" % datastream_namespace):
        if "cpe-oval" in component_el.get("id", ""):
            return component_el
    return None


def add_oval_definition_to_cpe_oval(root, unlinked_oval_file_path, oval_def_id):
    oval_cpe_root = get_cpe_oval_root(root)
    if oval_cpe_root is None:
        raise Exception("CPE OVAL is missing in base DS!")

    oval_document = get_linked_cpe_oval_document(unlinked_oval_file_path)

    references_to_keep = oval_document.get_all_references_of_definition(oval_def_id)
    oval_document.keep_referenced_components(references_to_keep)

    add_oval_components_to_oval_xml(
        oval_cpe_root, "definitions", oval_document.definitions
    )
    add_oval_components_to_oval_xml(oval_cpe_root, "tests", oval_document.tests)
    add_oval_components_to_oval_xml(oval_cpe_root, "objects", oval_document.objects)
    add_oval_components_to_oval_xml(oval_cpe_root, "states", oval_document.states)
    add_oval_components_to_oval_xml(oval_cpe_root, "variables", oval_document.variables)


def add_notice(benchmark, namespace, notice, warning):
    """
    Adds derivative notice as the first notice to given benchmark.
    """

    index = -1
    prev_element = None
    existing_notices = list(benchmark.findall("./{%s}notice" % (namespace)))
    if existing_notices:
        prev_element = existing_notices[0]
        # insert before the first notice
        index = list(benchmark).index(prev_element)
    else:
        existing_descriptions = list(
            benchmark.findall("./{%s}description" % (namespace))
        )
        prev_element = existing_descriptions[-1]
        # insert after the last description
        index = list(benchmark).index(prev_element) + 1

    if index == -1:
        raise RuntimeError(
            "Can't find existing notices or description in benchmark '%s'." %
            (benchmark)
        )

    elem = ElementTree.Element("{%s}notice" % (namespace))
    elem.set("id", warning)
    elem.append(notice)
    # this is done for the newline and indentation
    elem.tail = prev_element.tail
    benchmark.insert(index, elem)

    return True


def remove_idents(tree_root, namespace, prod="RHEL"):
    """
    Remove product identifiers from rules in XML tree
    """

    ident_exp = '.*' + prod + '-*'
    ref_exp = prod + '-*'
    for rule in tree_root.findall(".//{%s}Rule" % (namespace)):
        for ident in rule.findall(".//{%s}ident" % (namespace)):
            if ident is not None:
                if (re.search(r'CCE-*', ident.text) or
                        re.search(ident_exp, ident.text)):
                    rule.remove(ident)

        for ref in rule.findall(".//{%s}reference" % (namespace)):
            if ref.text is not None:
                if re.search(ref_exp, ref.text):
                    rule.remove(ref)

        for fix in rule.findall(".//{%s}fix" % (namespace)):
            sub_elems = fix.findall(".//{%s}sub" % (namespace))
            for sub_elem in sub_elems:
                sub_elem.tail = re.sub(r"[\s]+- CCE-.*", "", sub_elem.tail)
                sub_elem.tail = re.sub(r"CCE-[0-9]*-[0-9]*", "", sub_elem.tail)
            if fix.text is not None:
                fix.text = re.sub(r"[\s]+- CCE-.*", "", fix.text)
                fix.text = re.sub(r"CCE-[0-9]*-[0-9]*", "", fix.text)


def remove_cce_reference(tree_root, namespace):
    """
    Remove CCE identifiers from OVAL checks in XML tree
    """
    for definition in tree_root.findall(".//{%s}definition" % (namespace)):
        for metadata in definition.findall(".//{%s}metadata" % (namespace)):
            for ref in metadata.findall(".//{%s}reference" % (namespace)):
                if (re.search(r'CCE-*', ref.get("ref_id"))):
                    metadata.remove(ref)


def profile_handling(tree_root, namespace):
    ns_profiles = []
    for i in standard_profiles:
        ns_profiles.append("xccdf_%s.content_profile_%s" % (OSCAP_VENDOR, i))
    all_profiles = standard_profiles + ns_profiles
    for profile in tree_root.findall(".//{%s}Profile" % (namespace)):
        if profile.get("id") not in all_profiles:
            tree_root.remove(profile)


def replace_platform(tree_root, namespace, product):
    for oval in tree_root.findall(".//{%s}oval_definitions" % (namespace)):
        for platform in oval.findall(".//{%s}platform" % (namespace)):
            platform.text = (platform.text).replace("Red Hat Enterprise Linux", product)



import collections
import os
import re

import ssg.build_yaml
import ssg.constants

import template_renderer


def process_refs(ref_format, relevant_refs):
    matching_relevant_refs = []
    not_matching_relevant_refs = []

    for ref in relevant_refs:
        match = re.match(ref_format, ref)
        if match:
            matching_relevant_refs.append(ref)
        else:
            not_matching_relevant_refs.append(ref)

    sorted_relevant_refs = matching_relevant_refs + not_matching_relevant_refs
    return sorted_relevant_refs


def shorten_relevant_ref(ref_format, ref):
    match = re.match(ref_format, ref)
    if match:
        groups = []
        for refpart in match.groups():
            if refpart is None:
                refpart = ""
            try:
                refpart = "{0:07d}".format(int(refpart))
            except ValueError:
                pass
            groups.append(refpart)
        return tuple(groups)
    else:
        return TableHtmlOutput.DEFAULT_SHORTENED_REF


class TableHtmlOutput(template_renderer.Renderer):
    DEFAULT_SHORTENED_REF = ("~",)

    def __init__(self, * args, ** kwargs):
        super(TableHtmlOutput, self).__init__(* args, ** kwargs)

        self.rules_root = os.path.join(self.built_content_path, "rules")
        self.var_root = os.path.join(self.built_content_path, "values")

    def _get_var_value(self, varname):
        return self._get_var_value_from_default(varname)

    def _get_var_value_from_default(self, varname):
        var_path = os.path.join(self.var_root, varname + ".yml")
        var = ssg.build_yaml.Value.from_yaml(var_path, self.env_yaml)
        return var.options["default"]

    def _fix_var_sub_in_text(self, text, varname, value):
        return re.sub(
            r'<sub\s+idref="{var}"\s*/>'.format(var=varname),
            r"<tt>{val}</tt>".format(val=value), text)

    def _resolve_var_substitutions(self, rule):
        # The <sub .../> here is not the HTML subscript element <sub>...</sub>,
        # and therefore is invalid HTML.
        # so this code substitutes the whole sub element with contents of its idref prefixed by $
        # as occurrence of sub with idref implies that substitution of XCCDF values takes place
        variables = re.findall(r'<sub\s+idref="([^"]*)"\s*/>', rule.description)
        variables = set(variables)
        rule.substitutions = dict()
        for var in variables:
            val = self._get_var_value(var)
            rule.description = self._fix_var_sub_in_text(rule.description, var, val)
            rule.substitutions[var] = val

    def _get_eligible_rules(self, refcat):
        raise NotImplementedError

    def _generate_shortened_ref(self, reference, rule):
        if not rule.relevant_refs:
            return self.DEFAULT_SHORTENED_REF
        shortened_ref = shorten_relevant_ref(reference.regex_with_groups, rule.relevant_refs[0])
        if not shortened_ref:
            shortened_ref = self.DEFAULT_SHORTENED_REF
        return shortened_ref

    def process_rules(self, reference):
        eligible_rules = self._get_eligible_rules(reference.id)

        output_rules = collections.defaultdict(list)
        for rule in eligible_rules:
            self._resolve_var_substitutions(rule)
            relevant_refs = rule.references.get(reference.id, "")
            rule.relevant_refs = process_refs(reference.regex_with_groups, relevant_refs)
            shortened_ref = self._generate_shortened_ref(reference, rule)
            output_rules[shortened_ref].append(rule)

        self.template_data["rules_by_shortref"] = output_rules
        self.template_data["sorted_refs"] = sorted(list(output_rules.keys()))
        self.template_data["reference_title"] = reference.name
        self.template_data["product"] = self.product
        self.template_data["product_full_name"] = self.product
        for full, short in ssg.constants.FULL_NAME_TO_PRODUCT_MAPPING.items():
            if short == self.product:
                self.template_data["product_full_name"] = full
                break


def update_parser(parser):
    parser.add_argument(
        "refcategory", metavar="REFERENCE_ID",
        choices=ssg.constants.REFERENCES.keys(), help="Category of the rule reference")



import os
import pytest
import ssg.jinja


def get_definitions_with_substitution(subs_dict=None):
    if subs_dict is None:
        subs_dict = dict()

    definitions = os.path.join(os.path.dirname(__file__), "data", "definitions.jinja")
    ssg.jinja.update_substitutions_dict(definitions, subs_dict)
    return subs_dict


def test_macro_presence():
    actual_defs = get_definitions_with_substitution()
    assert "expand_to_bar" in actual_defs
    assert actual_defs["expand_to_bar"]() == "bar"


def test_macro_expansion():
    incomplete_defs = get_definitions_with_substitution()
    assert incomplete_defs["expand_to_global_var"]() == ""

    complete_defs = get_definitions_with_substitution(dict(global_var="value"))
    assert complete_defs["expand_to_global_var"]() == "value"


def test_load_macros_with_valid_directory(tmpdir):
    macros_dir = tmpdir.mkdir("macros")
    macro_file = macros_dir.join("test_macro.jinja")
    macro_file.write("{{% macro test_macro() %}}test{{% endmacro %}}")
    substitutions_dict = ssg.jinja._load_macros(str(macros_dir))

    assert "test_macro" in substitutions_dict
    assert substitutions_dict["test_macro"]() == "test"


def test_load_macros_with_nonexistent_directory():
    non_existent_dir = "/non/existent/directory"
    with pytest.raises(RuntimeError, match=f"The directory '{non_existent_dir}' does not exist."):
        ssg.jinja._load_macros(non_existent_dir)


def test_load_macros_with_existing_substitutions_dict(tmpdir):
    macros_dir = tmpdir.mkdir("macros")
    macro_file = macros_dir.join("test_macro.jinja")
    macro_file.write("{{% macro test_macro() %}}test{{% endmacro %}}")
    existing_dict = {"existing_key": "existing_value"}
    substitutions_dict = ssg.jinja._load_macros(str(macros_dir), existing_dict)

    assert "test_macro" in substitutions_dict
    assert substitutions_dict["test_macro"]() == "test"
    assert "existing_key" in substitutions_dict
    assert substitutions_dict["existing_key"] == "existing_value"



"""
Common functions for processing Environment in SSG
"""

from __future__ import absolute_import
from __future__ import print_function


from .products import load_product_yaml
from .yaml import open_raw


def open_environment(build_config_yaml_path, product_yaml_path, product_properties_path=None):
    """
    Opens and merges environment configurations from given YAML files.

    Args:
        build_config_yaml_path (str): The file path to the build configuration YAML file.
        product_yaml_path (str): The file path to the product YAML file.
        product_properties_path (str, optional): The directory path containing product properties
                                                 files. Defaults to None.

    Returns:
        dict: A dictionary containing the merged contents of the build configuration and product
              YAML files.
    """
    contents = open_raw(build_config_yaml_path)
    product = load_product_yaml(product_yaml_path)
    if product_properties_path:
        product.read_properties_from_directory(product_properties_path)
    contents.update(product)
    return contents



import os
import pytest

import ssg.components

ssg_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
data_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
components_dir = os.path.join(data_dir, "components_dir")
component_file = os.path.join(components_dir, "fapolicyd.yml")


def test_load():
    cs = ssg.components.load(components_dir)
    assert isinstance(cs, dict)
    assert len(cs) == 1
    assert "fapolicyd" in cs
    assert isinstance(cs["fapolicyd"], ssg.components.Component)
    assert cs["fapolicyd"].name == "fapolicyd"


def test_package_component_mapping():
    cs = ssg.components.load(components_dir)
    package_to_component = ssg.components.package_component_mapping(cs)
    assert isinstance(package_to_component, dict)
    assert len(package_to_component.keys()) == 1
    assert "fapolicyd-server" in package_to_component
    assert package_to_component["fapolicyd-server"] == ["fapolicyd"]


def test_template_component_mapping():
    cs = ssg.components.load(components_dir)
    template_to_component = ssg.components.template_component_mapping(cs)
    assert isinstance(template_to_component, dict)
    assert len(template_to_component.keys()) == 1
    assert "file_policy_blocked" in template_to_component
    assert template_to_component["file_policy_blocked"] == ["fapolicyd"]


def test_group_components_mapping():
    cs = ssg.components.load(components_dir)
    group_to_component = ssg.components.group_component_mapping(cs)
    assert isinstance(group_to_component, dict)
    assert len(group_to_component.keys()) == 2
    assert "fapolicy" in group_to_component
    assert len(group_to_component["fapolicy"]) == 1
    assert group_to_component["fapolicy"][0] == "fapolicyd"
    assert "integrity" in group_to_component
    assert len(group_to_component["integrity"]) == 1
    assert group_to_component["integrity"][0] == "fapolicyd"


def test_component_parse():
    c = ssg.components.Component(component_file)
    assert c.name == "fapolicyd"
    assert len(c.groups) == 2
    assert "fapolicy" in c.groups
    assert "integrity" in c.groups
    assert len(c.packages) == 1
    assert "fapolicyd-server" in c.packages
    assert len(c.rules) == 3
    assert "fapolicy_default_deny" in c.rules
    assert "fapolicyd_prevent_home_folder_access" in c.rules
    assert "service_fapolicyd_enabled" in c.rules
    assert len(c.templates) == 1
    assert "file_policy_blocked" in c.templates
    assert len(c.changelog) == 1
    assert "created in 2023" in c.changelog
